/*
    File:               argonAdvapi32Dll.cpp
    Description:        Dynamical loading of some functions of the
                        system's ADVAPI32.DLL.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         26.11.2005
    last updated:       09.02.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonAdvapi32Dll_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonAdvapi32Dll.h"

#ifndef _argon_no_includes
    #include "argonLocking.h"
#endif

namespace ArgonLibrary
{




BOOL WINAPI argonOpenProcessToken(
    IN HANDLE ProcessHandle,
    IN DWORD DesiredAccess,
    OUT PHANDLE TokenHandle
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnOpenProcessToken ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnOpenProcessToken( ProcessHandle, DesiredAccess, TokenHandle );
}

BOOL WINAPI argonLookupAccountSidA(
    IN LPCSTR lpSystemName,
    IN PSID Sid,
    OUT LPSTR Name,
    IN OUT LPDWORD cbName,
    OUT LPSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnLookupAccountSidA ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnLookupAccountSidA( lpSystemName,
                                                Sid,
                                                Name,
                                                cbName,
                                                ReferencedDomainName,
                                                cbReferencedDomainName,
                                                peUse );
}

BOOL WINAPI argonLookupAccountSidW(
    IN LPCWSTR lpSystemName,
    IN PSID Sid,
    OUT LPWSTR Name,
    IN OUT LPDWORD cbName,
    OUT LPWSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnLookupAccountSidW ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnLookupAccountSidW( lpSystemName,
                                                Sid,
                                                Name,
                                                cbName,
                                                ReferencedDomainName,
                                                cbReferencedDomainName,
                                                peUse );
}

BOOL WINAPI argonLookupAccountNameA(
    IN LPCSTR lpSystemName,
    IN LPCSTR lpAccountName,
    OUT PSID Sid,
    IN OUT LPDWORD cbSid,
    OUT LPSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnLookupAccountNameA ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnLookupAccountNameA( lpSystemName,
                                                 lpAccountName,
                                                 Sid,
                                                 cbSid,
                                                 ReferencedDomainName,
                                                 cbReferencedDomainName,
                                                 peUse );
}

BOOL WINAPI argonLookupAccountNameW(
    IN LPCWSTR lpSystemName,
    IN LPCWSTR lpAccountName,
    OUT PSID Sid,
    IN OUT LPDWORD cbSid,
    OUT LPWSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnLookupAccountNameW ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnLookupAccountNameW( lpSystemName,
                                                 lpAccountName,
                                                 Sid,
                                                 cbSid,
                                                 ReferencedDomainName,
                                                 cbReferencedDomainName,
                                                 peUse );
}

BOOL WINAPI argonLookupPrivilegeValueA(
    IN LPCSTR lpSystemName,
    IN LPCSTR lpName,
    OUT PLUID lpLuid
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnLookupPrivilegeValueA ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnLookupPrivilegeValueA( lpSystemName, lpName, lpLuid );
}

BOOL WINAPI argonLookupPrivilegeValueW(
    IN LPCWSTR lpSystemName,
    IN LPCWSTR lpName,
    OUT PLUID lpLuid
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnLookupPrivilegeValueW ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnLookupPrivilegeValueW( lpSystemName, lpName, lpLuid );
}

BOOL WINAPI argonGetTokenInformation(
    IN HANDLE TokenHandle,
    IN TOKEN_INFORMATION_CLASS TokenInformationClass,
    OUT LPVOID TokenInformation,
    IN DWORD TokenInformationLength,
    OUT PDWORD ReturnLength
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnGetTokenInformation ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return Advapi32Dll::s_pfnGetTokenInformation( TokenHandle,
                                                  TokenInformationClass,
                                                  TokenInformation,
                                                  TokenInformationLength,
                                                  ReturnLength );
}

BOOL WINAPI argonAdjustTokenPrivileges(
    IN HANDLE TokenHandle,
    IN BOOL DisableAllPrivileges,
    IN PTOKEN_PRIVILEGES NewState,
    IN DWORD BufferLength,
    OUT PTOKEN_PRIVILEGES PreviousState,
    OUT PDWORD ReturnLength
)
{
    Advapi32Dll::LoadAPI();
    if(   !Advapi32Dll::s_pfnAdjustTokenPrivileges ) { SetLastError( ERROR_NOT_ALL_ASSIGNED ); return FALSE; }
    return Advapi32Dll::s_pfnAdjustTokenPrivileges( TokenHandle,
                                                    DisableAllPrivileges,
                                                    NewState,
                                                    BufferLength,
                                                    PreviousState,
                                                    ReturnLength );
}




Advapi32Dll::Unloader    Advapi32Dll::s_Unloader;
HMODULE                  Advapi32Dll::s_hModule                  = NULL;
TfnOpenProcessToken      Advapi32Dll::s_pfnOpenProcessToken      = NULL;
TfnLookupAccountSidA     Advapi32Dll::s_pfnLookupAccountSidA     = NULL;
TfnLookupAccountSidW     Advapi32Dll::s_pfnLookupAccountSidW     = NULL;
TfnLookupAccountNameA    Advapi32Dll::s_pfnLookupAccountNameA    = NULL;
TfnLookupAccountNameW    Advapi32Dll::s_pfnLookupAccountNameW    = NULL;
TfnLookupPrivilegeValueA Advapi32Dll::s_pfnLookupPrivilegeValueA = NULL;
TfnLookupPrivilegeValueW Advapi32Dll::s_pfnLookupPrivilegeValueW = NULL;
TfnGetTokenInformation   Advapi32Dll::s_pfnGetTokenInformation   = NULL;
TfnAdjustTokenPrivileges Advapi32Dll::s_pfnAdjustTokenPrivileges = NULL;
bool                     Advapi32Dll::s_bAutoUnload              = true;

Advapi32Dll::Advapi32Dll()
{
}

// Remarks: We don't need any object of #Advapi32Dll#.
//  We just need its static members.



bool Advapi32Dll::hasAPI()
{
    LoadAPI();
    return s_pfnOpenProcessToken != NULL;
    // IMPORTANT: The "ADVAPI32.DLL" is always loadable because it contains windows functions
    //  for already very early Windows versions. So we must check separately if our
    //  security-related-functions are included in the dll or not.
}



void Advapi32Dll::LoadAPI()
{
    if( s_hModule ) return;

    LockModuleWide lmw;

    if( s_hModule ) return;

    Nullify();

    s_hModule = ::LoadLibrary( _T(ADVAPI32DLL_NAME) );

    if( s_hModule )
    {
        s_pfnOpenProcessToken      = (TfnOpenProcessToken     ) GetProcAddress( s_hModule, "OpenProcessToken"      );
        s_pfnLookupAccountSidA     = (TfnLookupAccountSidA    ) GetProcAddress( s_hModule, "LookupAccountSidA"     );
        s_pfnLookupAccountSidW     = (TfnLookupAccountSidW    ) GetProcAddress( s_hModule, "LookupAccountSidW"     );
        s_pfnLookupAccountNameA    = (TfnLookupAccountNameA   ) GetProcAddress( s_hModule, "LookupAccountNameA"    );
        s_pfnLookupAccountNameW    = (TfnLookupAccountNameW   ) GetProcAddress( s_hModule, "LookupAccountNameW"    );
        s_pfnLookupPrivilegeValueA = (TfnLookupPrivilegeValueA) GetProcAddress( s_hModule, "LookupPrivilegeValueA" );
        s_pfnLookupPrivilegeValueW = (TfnLookupPrivilegeValueW) GetProcAddress( s_hModule, "LookupPrivilegeValueW" );
        s_pfnGetTokenInformation   = (TfnGetTokenInformation  ) GetProcAddress( s_hModule, "GetTokenInformation"   );
        s_pfnAdjustTokenPrivileges = (TfnAdjustTokenPrivileges) GetProcAddress( s_hModule, "AdjustTokenPrivileges" );
    }
}

void Advapi32Dll::UnloadAPI()
{
    LockModuleWide lmw;

    if( s_hModule )
        ::FreeLibrary( s_hModule );

    Nullify();
}



void Advapi32Dll::Nullify()
{
    s_pfnOpenProcessToken      = NULL;
    s_pfnLookupAccountSidA     = NULL;
    s_pfnLookupAccountSidW     = NULL;
    s_pfnLookupAccountNameA    = NULL;
    s_pfnLookupAccountNameW    = NULL;
    s_pfnLookupPrivilegeValueA = NULL;
    s_pfnLookupPrivilegeValueW = NULL;
    s_pfnGetTokenInformation   = NULL;
    s_pfnAdjustTokenPrivileges = NULL;
    s_hModule = NULL;
}




}
