/*
    File:               argonMoreMessageBox.cpp
    Description:        Provides a class with an enhanced messagebox which
                        includes additional fields and icons and a more button.

    Author:             Roman Komary
    created on:         15.02.2006
    last updated:       06.10.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary. All rights reserved.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)

    Thanks go to Jovita De Guzman for the italian translations.
*/

#define __argonMoreMessageBox_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonMoreMessageBox.h"

namespace ArgonLibrary
{




// Language Table strings for the MoreMessageBox (Buttons etc.).

const LangTableDesc g_MMBLangDescs[] = {
    #ifndef _argon_no_LangEnglish
    // English
    { mmbid_Ok,                     LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&OK") },
    { mmbid_Cancel,                 LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Cancel") },
    { mmbid_Yes,                    LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Yes") },
    { mmbid_No,                     LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&No") },
    { mmbid_Abort,                  LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Abort") },
    { mmbid_Retry,                  LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Retry") },
    { mmbid_Ignore,                 LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Ignore") },
    { mmbid_TryAgain,               LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Try Again") },
    { mmbid_Continue,               LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Continue") },
    { mmbid_Help,                   LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&Help") },
    { mmbid_More__,                 LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("&More ...") },
    { mmbid_AlwaysShowThisMessage,  LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("Always &show this message") },
    { mmbid_Information,            LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("Information") },
    #endif

    #ifndef _argon_no_LangGerman
    // German
    { mmbid_Ok,                     LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&OK") },
    { mmbid_Cancel,                 LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Abbrechen") },
    { mmbid_Yes,                    LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Ja") },
    { mmbid_No,                     LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Nein") },
    { mmbid_Abort,                  LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Abbrechen") },
    { mmbid_Retry,                  LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Wiederholen") },
    { mmbid_Ignore,                 LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Ignorieren") },
    { mmbid_TryAgain,               LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("Wie&derholen") },
    { mmbid_Continue,               LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Weiter") },
    { mmbid_Help,                   LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Hilfe") },
    { mmbid_More__,                 LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Mehr ...") },
    { mmbid_AlwaysShowThisMessage,  LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("&Diese Nachricht immer anzeigen") },
    { mmbid_Information,            LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("Information") },
    #endif

    #ifndef _argon_no_LangItalian
    // Italian
    { mmbid_Ok,                     LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&OK") },
    { mmbid_Cancel,                 LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Cancella") },
    { mmbid_Yes,                    LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Si") },
    { mmbid_No,                     LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&No") },
    { mmbid_Abort,                  LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Termina") },
    { mmbid_Retry,                  LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Riprova") },
    { mmbid_Ignore,                 LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Ignora") },
    { mmbid_TryAgain,               LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Riprova") },
    { mmbid_Continue,               LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Continua") },
    { mmbid_Help,                   LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Aiuto") },
    { mmbid_More__,                 LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Pi Info ...") },
    { mmbid_AlwaysShowThisMessage,  LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("&Mostra sempre questo messaggio.") },
    { mmbid_Information,            LANG_ITALIAN,   ltmapt_Direct,  (LONG_PTR) _T("Informazione") },
    #endif

    LANGTABLEDESC_ENDMARK
};

MMBLanguageTable                            g_MMBLanguageTable;
LanguageDescriptionTableT<MMBLanguageTable> g_MMBLanguageDescriptionTable( g_MMBLangDescs );




#ifndef _WIN32_WCE

// DIALOGTEMPLATEEX structure and DLGITEMTEMPLATEEX items
// building the MoreMessageBox dialog.
static const BYTE s_abyMoreMessageBox[] =
{
    0x01, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC8, 0x08, 0xC8, 0x80, // ...........Ȁ
    0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x98, 0x01, 0x19, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // ...............
    0x08, 0x00, 0x90, 0x01, 0x00, 0x01, 0x4D, 0x00, 0x53, 0x00, 0x20, 0x00, 0x53, 0x00, 0x68, 0x00, // .....M.S. .S.h.
    0x65, 0x00, 0x6C, 0x00, 0x6C, 0x00, 0x20, 0x00, 0x44, 0x00, 0x6C, 0x00, 0x67, 0x00, 0x00, 0x00, // e.l.l. .D.l.g...
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x03, 0x50, 0x09, 0x00, 0xC2, 0x00, // ...........P...
    0x3C, 0x00, 0x0E, 0x00, 0xF0, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x42, 0x00, 0x75, 0x00, // <.......B.u.
    0x74, 0x00, 0x74, 0x00, 0x6F, 0x00, 0x6E, 0x00, 0x31, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // t.t.o.n.1.......
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x50, 0x4B, 0x00, 0xC2, 0x00, // ...........PK..
    0x3C, 0x00, 0x0E, 0x00, 0xF1, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x42, 0x00, 0x75, 0x00, // <.......B.u.
    0x74, 0x00, 0x74, 0x00, 0x6F, 0x00, 0x6E, 0x00, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // t.t.o.n.2.......
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x50, 0x8D, 0x00, 0xC2, 0x00, // ...........P..
    0x3C, 0x00, 0x0E, 0x00, 0xF2, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x42, 0x00, 0x75, 0x00, // <.......B.u.
    0x74, 0x00, 0x74, 0x00, 0x6F, 0x00, 0x6E, 0x00, 0x33, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // t.t.o.n.3.......
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x50, 0xCF, 0x00, 0xC2, 0x00, // ...........P..
    0x3C, 0x00, 0x0E, 0x00, 0xF6, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x42, 0x00, 0x75, 0x00, // <.......B.u.
    0x74, 0x00, 0x74, 0x00, 0x6F, 0x00, 0x6E, 0x00, 0x34, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // t.t.o.n.4.......
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x50, 0x11, 0x01, 0xC2, 0x00, // ...........P...
    0x3C, 0x00, 0x0E, 0x00, 0xF7, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x26, 0x00, 0x48, 0x00, // <.......&.H.
    0x65, 0x00, 0x6C, 0x00, 0x70, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // e.l.p...........
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x50, 0x53, 0x01, 0xC2, 0x00, 0x3C, 0x00, 0x0E, 0x00, // .......PS..<...
    0xF3, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x26, 0x00, 0x4D, 0x00, 0x6F, 0x00, 0x72, 0x00, // ....&.M.o.r.
    0x65, 0x00, 0x20, 0x00, 0x2E, 0x00, 0x2E, 0x00, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // e. .............
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x50, 0x09, 0x00, 0xD9, 0x00, // ...........P...
    0x86, 0x01, 0x0A, 0x00, 0xF4, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x43, 0x00, 0x68, 0x00, // .......C.h.
    0x65, 0x00, 0x63, 0x00, 0x6B, 0x00, 0x31, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // e.c.k.1.........
    0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x01, 0x50, 0x09, 0x00, 0xE7, 0x00, 0x86, 0x01, 0x0A, 0x00, // .......P......
    0xF5, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x80, 0x00, 0x43, 0x00, 0x68, 0x00, 0x65, 0x00, 0x63, 0x00, // ....C.h.e.c.
    0x6B, 0x00, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // k.2.............
    0x10, 0x00, 0x00, 0x50, 0x00, 0x00, 0xF7, 0x00, 0x98, 0x01, 0x01, 0x00, 0xFC, 0x03, 0x00, 0x00, // ...P.........
    0xFF, 0xFF, 0x82, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // .............
    0x03, 0x00, 0x00, 0x50, 0x09, 0x00, 0xFE, 0x00, 0x14, 0x00, 0x14, 0x00, 0xF9, 0x03, 0x00, 0x00, // ...P..........
    0xFF, 0xFF, 0x82, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // .............
    0x80, 0x00, 0x02, 0x50, 0x2B, 0x00, 0xFE, 0x00, 0x64, 0x01, 0x0B, 0x00, 0xFA, 0x03, 0x00, 0x00, // ..P+..d......
    0xFF, 0xFF, 0x82, 0x00, 0x53, 0x00, 0x74, 0x00, 0x61, 0x00, 0x74, 0x00, 0x69, 0x00, 0x63, 0x00, // .S.t.a.t.i.c.
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x01, 0x02, 0x50, // ..............P
    0x2B, 0x00, 0x0A, 0x01, 0x64, 0x01, 0x08, 0x00, 0xFB, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x82, 0x00, // +...d.......
    0x53, 0x00, 0x74, 0x00, 0x61, 0x00, 0x74, 0x00, 0x69, 0x00, 0x63, 0x00, 0x00, 0x00, 0x00, 0x00, // S.t.a.t.i.c.....
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x02, 0x50, 0x09, 0x00, 0x07, 0x00, // ..........P....
    0x86, 0x01, 0x08, 0x00, 0xF8, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x82, 0x00, 0x53, 0x00, 0x74, 0x00, // .......S.t.
    0x61, 0x00, 0x74, 0x00, 0x69, 0x00, 0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // a.t.i.c.........
    0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x50, 0x09, 0x00, 0x18, 0x00, 0x14, 0x00, 0x14, 0x00, // .......P........
    0xED, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x82, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // ............
    0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x02, 0x50, 0x2B, 0x00, 0x18, 0x00, 0x64, 0x01, 0x14, 0x00, // ......P+...d...
    0xEE, 0x03, 0x00, 0x00, 0xFF, 0xFF, 0x82, 0x00, 0x53, 0x00, 0x74, 0x00, 0x61, 0x00, 0x74, 0x00, // ....S.t.a.t.
    0x69, 0x00, 0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // i.c.............
    0xC4, 0x08, 0xB1, 0x50, 0x2B, 0x00, 0x37, 0x00, 0x64, 0x01, 0x7F, 0x00, 0xEF, 0x03, 0x00, 0x00, // .P+.7.d.....
    0xFF, 0xFF, 0x81, 0x00, 0x00, 0x00, 0x00, 0x00                                                  // .....
};

/*
The above dialog template is the binary result of a resource
dialog that looks like this:

mmbIDD_MOREMESSAGEBOX DIALOGEX 0, 0, 408, 281
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | DS_CENTER | WS_POPUP | 
    WS_CAPTION | WS_SYSMENU
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "Button1",mmbIDC_BUTTON1,9,194,60,14,WS_GROUP
    PUSHBUTTON      "Button2",mmbIDC_BUTTON2,75,194,60,14
    PUSHBUTTON      "Button3",mmbIDC_BUTTON3,141,194,60,14
    PUSHBUTTON      "Button4",mmbIDC_BUTTON4,207,194,60,14
    PUSHBUTTON      "&Help",mmbIDC_HELPBTN,273,194,60,14
    PUSHBUTTON      "&More ...",mmbIDC_MORE,339,194,60,14
    CONTROL         "Check1",mmbIDC_CHECK1,"Button",BS_AUTOCHECKBOX | 
                    WS_GROUP | WS_TABSTOP,9,217,390,10
    CONTROL         "Check2",mmbIDC_CHECK2,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,9,231,390,10
    CONTROL         "",mmbIDC_HR,"Static",SS_ETCHEDHORZ,0,247,408,1
    ICON            "",mmbIDC_EXTRAICON,9,254,20,20
    LTEXT           "Static",mmbIDC_EXTRAMSG,43,254,356,11,SS_NOPREFIX
    LTEXT           "Static",mmbIDC_EXTRALINK,43,266,356,8,SS_NOPREFIX | 
                    SS_NOTIFY
    LTEXT           "Static",mmbIDC_HEADER,9,7,390,8,SS_NOPREFIX
    ICON            "",mmbIDC_MSGICON,9,24,20,20
    LTEXT           "Static",mmbIDC_MSG,43,24,356,20,SS_NOPREFIX
    EDITTEXT        mmbIDC_LONGMSG,43,55,356,127,ES_MULTILINE | 
                    ES_AUTOVSCROLL | ES_AUTOHSCROLL | ES_READONLY | 
                    WS_VSCROLL | WS_HSCROLL
END

with the following ID values:

#define mmbIDD_MOREMESSAGEBOX           132
#define mmbIDC_MSGICON                  1005
#define mmbIDC_MSG                      1006
#define mmbIDC_LONGMSG                  1007
#define mmbIDC_BUTTON1                  1008
#define mmbIDC_BUTTON2                  1009
#define mmbIDC_BUTTON3                  1010
#define mmbIDC_MORE                     1011
#define mmbIDC_CHECK1                   1012
#define mmbIDC_CHECK2                   1013
#define mmbIDC_BUTTON4                  1014
#define mmbIDC_HELPBTN                  1015
#define mmbIDC_HEADER                   1016
#define mmbIDC_EXTRAICON                1017
#define mmbIDC_EXTRAMSG                 1018
#define mmbIDC_EXTRALINK                1019
#define mmbIDC_HR                       1020
*/

/*
    And following the bitmaps for the up and down arrow icons that look
    something like this (in black):

            x
           xxx
          xxxxx
         xxxxxxx
        xxxxxxxxx
          xxxxx
          xxxxx


          xxxxx
          xxxxx
        xxxxxxxxx
         xxxxxxx
          xxxxx
           xxx
            x
*/

// monochrome bitmap AND mask for the up arrow icon
static const BYTE s_abyUpArrowBitmap9x7_AndMask[] =
{
    0xF7, 0x80,
    0xE3, 0x80,
    0xC1, 0x80,
    0x80, 0x80,
    0x00, 0x00,
    0xC1, 0x80,
    0xC1, 0x80
};

// monochrome bitmap AND mask for the down arrow icon
static const BYTE s_abyDownArrowBitmap9x7_AndMask[] =
{
    0xC1, 0x80,
    0xC1, 0x80,
    0x00, 0x00,
    0x80, 0x80,
    0xC1, 0x80,
    0xE3, 0x80,
    0xF7, 0x80
};

// monochrome bitmap XOR mask for the down arrow icon
static const BYTE s_abyUpDownArrowBitmap9x7_XorMask[] =
{
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00
};

#endif

#pragma warning( disable : 4244 )       // warning C4244: 'argument' : conversion from 'LONG_PTR' to 'LONG', possible loss of data
#pragma warning( disable : 4312 )       // warning C4312: 'type cast' : conversion from 'LONG' to 'ArgonLibrary::MoreMessageBox *' of greater size

MoreMessageBox::MoreMessageBox()
{
    #ifndef _WIN32_WCE
    m_minWidthAsNumButtons = 0;
    m_rectInflateBorders.left   =
    m_rectInflateBorders.top    =
    m_rectInflateBorders.right  =
    m_rectInflateBorders.bottom = 0;

    m_bIs_Check1_Checked = true;
    m_bIs_Check2_Checked = false;
    m_idMessage_UrlClicked = 0;

    m_sizeUpDownArrow.cx = 9;
    m_sizeUpDownArrow.cy = 7;
    m_hiconUpArrow   = CreateIcon( Get_Module_Instance(),
                                   m_sizeUpDownArrow.cx, m_sizeUpDownArrow.cy,
                                   1, 1,
                                   s_abyUpArrowBitmap9x7_AndMask,
                                   s_abyUpDownArrowBitmap9x7_XorMask );
    m_hiconDownArrow = CreateIcon( Get_Module_Instance(),
                                   m_sizeUpDownArrow.cx, m_sizeUpDownArrow.cy,
                                   1, 1,
                                   s_abyDownArrowBitmap9x7_AndMask,
                                   s_abyUpDownArrowBitmap9x7_XorMask );
    m_nOffsetUpDownArrow = 6;
    #endif
}

MoreMessageBox::~MoreMessageBox()
{
    #ifndef _WIN32_WCE
    if( m_hiconDownArrow )
        { DestroyIcon( m_hiconDownArrow ); m_hiconDownArrow = NULL; }
    #endif
}

///<summary>
///  <b>MessageBox()</b> is similar to the system's MessageBox function,
///   but it contains an additional "More" button which, when pressed,
///   will display a long message text additional to the normal message
///   text. <br/>
///  Restriction: Only Ok button (MB_OK) is supported. But all MB_ICON*
///   icons are supported. <br/>
///  In addition, the width of the messagebox will not be adjusted,
///   only its height. <br/>
///  Beside the MB_ICON* flags, only the following flags will have effect:
///   All MB_TYPEMASK contained flags (MB_OK, MB_YESNO, etc.),
///   all MB_DEFBUTTON* flags,
///   MB_APPLMODAL, MB_SYSTEMMODAL, MB_SETFOREGROUND, MB_TOPMOST.
///</summary>
///<param name="hwndParent"> parent window </param>
///<param name="pstrMsg"> normal message text </param>
///<param name="pstrLongMsg"> long message text displayed when the
///  "More" button has been pressed. </param>
///<param name="pstrCaption"> dialog caption text </param>
///<param name="uType"> combination of MB_* values. Similar to the system's
///   MessageBox function. <br/>
///  Note: You will also hear the sound defined for an MB_ICON* flag
///   but only if a sound is defined and only if System-Beeps are enabled
///   in your system. </param>
///<param name="dwMmbFlags"> combination of MMB_* flags </param>
///<param name="language"> Define the desired language (locale) to use for the
///  MessageBox controls and caption texts. <br/>
///  Usually you would pass the locale returned by <b>GetThreadLocale()</b>. </param>
///<param name="pstrHeader"> Some text displayed above the <b>pstrMsg</b> text
///  and above the MsgIcon. Usually this text will be written using a bold font. <br/>
///  If NULL or empty string, no header will be displayed. </param>
///<param name="pstrExtraMsg"> Some text displayed in the bottom area (the so-called
///  Extra area). If NULL or empty string, no extra message is displayed. </param>
///<param name="pstrExtraLink_Text"> Below the extra message, a Link can be
///   displayed the user may click. This text will be written using an underline font. <br/>
///  If NULL or empty string, the text in <b>pstrExtraLink_Url</b> will be displayed
///   directly. <br/>
///  If both, <b>pstrExtraLink_Text</b> and <b>pstrExtraLink_Url</b> are NULL or empty
///   strings, no extra link will be displayed. </param>
///<param name="pstrExtraLink_Url"> The url (or commandline command) to execute
///   when the user clicks on it. <br/>
///  If both, <b>pstrExtraLink_Text</b> and <b>pstrExtraLink_Url</b> are NULL or empty
///   strings, no extra link will be displayed. </param>
///<param name="hModuleToLoadIconsFrom"> A module handle used to load icon resources
///  from. If NULL, the following icon parameters will be interpreted as HICONs
///  rather than resource icon IDs. </param>
///<param name="nMsgIcon_IdOrHicon"> ID or HICON of the icon to display left to the
///   main message text. <br/>
///  If <b>hModuleToLoadIconsFrom</b> is NULL, this indicates a HICON handle,
///   otherwise it indicates a resource icon ID to be loaded from the
///   <b>hModuleToLoadIconsFrom</b> module. <br/>
///  If this parameter is 0, no user-defined icon will be displayed. Only the icon
///   defined in the <b>uType</b> parameter will be displayed instead. <br/>
///  If this parameter is not 0, the icon will be displayed regardless if the <b>uType</b>
///   parameter contains a icon flag. In that case, the sound to the icon flag can be
///   heard, but the user-defined icon will be displayed. </param>
///<param name="nExtraIcon_IdOrHicon"> ID or HICON of the icon to display left to the
///   extra message text. <br/>
///  If <b>hModuleToLoadIconsFrom</b> is NULL, this indicates a HICON handle,
///   otherwise it indicates a resource icon ID to be loaded from the
///   <b>hModuleToLoadIconsFrom</b> module. <br/>
///  If this parameter is 0, no extra icon will be displayed. </param>
///<returns> IDOK if the ok button has been pressed. </returns>
INT_PTR MoreMessageBox::MessageBox( HWND hwndParent,
                                    PCTSTR pstrMsg, PCTSTR pstrLongMsg, PCTSTR pstrCaption,
                                    UINT uType, DWORD dwMmbFlags, LCID language,
                                    PCTSTR pstrHeader, PCTSTR pstrExtraMsg, PCTSTR pstrExtraLink_Text, PCTSTR pstrExtraLink_Url,
                                    HMODULE hModuleToLoadIconsFrom, UINT_PTR nMsgIcon_IdOrHicon, UINT_PTR nExtraIcon_IdOrHicon )
{
    #ifndef _WIN32_WCE

    m_aDisabledPopups.Empty();
    autoEmpty<Array<HWND>::typeArray> ae( &m_aDisabledPopups );

    m_Language = language;
    m_strMsg = pstrMsg;
    m_strLongMsg = pstrLongMsg;
    m_strLongMsg.MakeCRLF();
    m_strCaption = pstrCaption ? pstrCaption : MMBLanguageTable::getString(mmbid_Information, language);

    if( dwMmbFlags & MMB_AppendExeNameInCaption )
        m_strCaption += _T(" (") + String().GetExecutablefilename().FileTitleOf() + _T(')');

    if( !hwndParent )
        hwndParent = FindNiceParent();

    m_uType = uType;

    switch( uType & MB_ICONMASK )
    {
        case MB_ICONHAND        : m_uMsgIcon = (UINT_PTR) IDI_HAND;        break;
        case MB_ICONQUESTION    : m_uMsgIcon = (UINT_PTR) IDI_QUESTION;    break;
        case MB_ICONEXCLAMATION : m_uMsgIcon = (UINT_PTR) IDI_EXCLAMATION; break;
        case MB_ICONASTERISK    : m_uMsgIcon = (UINT_PTR) IDI_ASTERISK;    break;
        default                 : m_uMsgIcon = 0;                          break;
    }

    m_uMsgBeep      = uType & MB_ICONMASK;
    m_uMsgButtons   = uType & MB_TYPEMASK;
    m_uMsgDefButton = uType & MB_DEFMASK;
    m_dwMmbFlags    = dwMmbFlags;
    m_Controls.Empty();
    m_hwndDefBtn = NULL;

    HGLOBAL hMem = GlobalAlloc( GHND, sizeof(s_abyMoreMessageBox) );
    if( !hMem ) return -1;
    void *pMem = GlobalLock( hMem );
    if( !pMem ) { GlobalFree( hMem ); return -1; }

    memcpy( pMem, s_abyMoreMessageBox, sizeof(s_abyMoreMessageBox) );
    DWORD *pExStyle = (DWORD*) ((BYTE*) pMem +  8);
    DWORD *pStyle   = (DWORD*) ((BYTE*) pMem + 12);

    if( dwMmbFlags & MMB_AppWindow ) *pExStyle |= WS_EX_APPWINDOW;
    if( dwMmbFlags & MMB_ToolWindow ) *pExStyle |= WS_EX_TOOLWINDOW;
    if( uType & MB_TOPMOST ) *pExStyle |= WS_EX_TOPMOST;
    if( (uType & MB_MODEMASK) == MB_SYSTEMMODAL ) *pStyle |= DS_SYSMODAL;

    m_strHeaderMsg      = pstrHeader;
    m_strExtraMsg       = pstrExtraMsg;
    m_strExtraLink_Text = pstrExtraLink_Text;
    m_strExtraLink_Url  = pstrExtraLink_Url;
    if( m_strExtraLink_Text.IsEmpty() ) m_strExtraLink_Text = m_strExtraLink_Url;

    m_hModuleToLoadIconsFrom = hModuleToLoadIconsFrom;
    m_nMsgIcon_IdOrHicon     = nMsgIcon_IdOrHicon;
    m_nExtraIcon_IdOrHicon   = nExtraIcon_IdOrHicon;

    // The flags to be used for DrawText() by the MoreMessageBox.
    m_DT_MMB = (dwMmbFlags & MMB_TextLikeMessageBox) ?
        DT_EXPANDTABS : DT_EDITCONTROL | DT_WORDBREAK | DT_EXPANDTABS;
    m_DT_MMBHeader =
    m_DT_MMBExtra = DT_EDITCONTROL | DT_WORDBREAK | DT_EXPANDTABS;

    if( !(dwMmbFlags & MMB_DontDisablePopups) )
        GetOwnedPopups( hwndParent, m_aDisabledPopups, OwnedPopupsF_skipDisabledPopups );
        // Note: You could think we would run into problems regarding disabled
        // popups when we occasionally open 2 modal MoreMessageBoxes with the same parent window.
        // The 1st one disables all popups, but the 2nd does not do this again
        // because of the flag OwnedPopupsF_skipDisabledPopups.
        // You would think, if the 1st box is closed first, the popups are enabled
        // again while the 2nd box still is open.
        // But this cannot happen since when the 2nd box is created, it will
        // identify the 1st box as an owned popup since both have the same
        // parent window. Thus, the 2nd box will disable the 1st box.
        // So you *have* to close the 2nd box first, then you can close
        // the 1st one. So, we do not have a problem.

    EnableDisabledPopups( false );

    INT_PTR ret = DialogBoxIndirectParam( Get_Module_Instance(),
                                          (DLGTEMPLATE*) pMem,
                                          hwndParent,
                                          WndProc,
                                          (LPARAM) this );

    EnableDisabledPopups( true );

    GlobalUnlock( hMem );
    GlobalFree( hMem );
    return ret;

    #else

    return ::MessageBox( hwndParent, pstrMsg, pstrCaption, uType );

    #endif
}

#ifndef _WIN32_WCE

///<summary>
///  <b>onInitDialog()</b> is called when the messagebox dialog is being initialized.
///</summary>
///<returns> TRUE if the dialog should set the focus on a control,
///  FALSE if you did set the focus on a control yourself. </returns>
BOOL MoreMessageBox::onInitDialog()
{
    int minAdjustedWidthAsNumButtons = m_minWidthAsNumButtons;
    if( m_minWidthAsNumButtons == 0 )
    {
        if( ((m_dwMmbFlags & MMB_NoMoreButton) || m_strLongMsg.IsEmpty()) &&
            !m_uMsgIcon &&
            (m_dwMmbFlags & MMB_TextLikeMessageBox) &&
            (m_dwMmbFlags & MMB_Mask_NumCheckboxes) == 0 &&
            m_strExtraMsg.IsEmpty() && m_strExtraLink_Text.IsEmpty() )
        {
            minAdjustedWidthAsNumButtons = 1;
        }
        else
            minAdjustedWidthAsNumButtons = 4;
    }

    m_Controls.getFromDialog( m_hWnd,
                              m_dwMmbFlags, minAdjustedWidthAsNumButtons );

    HFONT hfontNormal = (HFONT) SendMessage( m_hWnd, WM_GETFONT, 0, 0 );
    m_hfontBold       = CreateAdoptedFont( hfontNormal, FW_BOLD );
    m_hfontUnderlined = CreateAdoptedFont( hfontNormal, FW_NORMAL, false, true );

    if( m_Controls.m_hwndHeaderMsg && m_hfontBold && !(m_dwMmbFlags & MMB_HeaderMsgNonBold) )
        SendMessage( m_Controls.m_hwndHeaderMsg, WM_SETFONT, (WPARAM) m_hfontBold, TRUE );
    if( m_Controls.m_hwndMsg && m_hfontBold && (m_dwMmbFlags & MMB_MsgBold) )
        SendMessage( m_Controls.m_hwndMsg, WM_SETFONT, (WPARAM) m_hfontBold, TRUE );
    if( m_Controls.m_hwndExtraMsg && m_hfontBold && (m_dwMmbFlags & MMB_ExtraMsgBold) )
        SendMessage( m_Controls.m_hwndExtraMsg, WM_SETFONT, (WPARAM) m_hfontBold, TRUE );
    if( m_Controls.m_hwndExtraLink && m_hfontUnderlined )
        SendMessage( m_Controls.m_hwndExtraLink, WM_SETFONT, (WPARAM) m_hfontUnderlined, TRUE );

    if( m_strHeaderMsg.IsEmpty() )
        m_Controls.deleteHeaderMsg();

    if( m_nMsgIcon_IdOrHicon )
    {
        HICON hicon = m_hModuleToLoadIconsFrom ?
            LoadIcon( m_hModuleToLoadIconsFrom, (PCTSTR) m_nMsgIcon_IdOrHicon ) : (HICON) m_nMsgIcon_IdOrHicon;
        SendMessage( m_Controls.m_hwndMsgIcon, STM_SETICON, (WPARAM) hicon, 0 );
        m_Controls.getMsgIconSize();
    }
    else
    if( m_uMsgIcon )
    {
        HICON hicon = LoadIcon( NULL, (PCTSTR) m_uMsgIcon );
        SendMessage( m_Controls.m_hwndMsgIcon, STM_SETICON, (WPARAM) hicon, 0 );
        m_Controls.getMsgIconSize();
    }
    else
        m_Controls.deleteMsgIcon();

    if( m_nExtraIcon_IdOrHicon )
    {
        HICON hicon = m_hModuleToLoadIconsFrom ?
            LoadIcon( m_hModuleToLoadIconsFrom, (PCTSTR) m_nExtraIcon_IdOrHicon ) : (HICON) m_nExtraIcon_IdOrHicon;
        SendMessage( m_Controls.m_hwndExtraIcon, STM_SETICON, (WPARAM) hicon, 0 );
        m_Controls.getExtraIconSize();
    }
    else
        m_Controls.deleteExtraArea( true, false, false );

    m_Controls.deleteExtraArea( false, m_strExtraMsg.IsEmpty(), m_strExtraLink_Text.IsEmpty() );

    int cButtons = 0;
    switch( m_uMsgButtons )
    {
        default :
        case MB_OK                : cButtons = 1; break;
        case MB_OKCANCEL          : cButtons = 2; break;
        case MB_ABORTRETRYIGNORE  : cButtons = 3; break;
        case MB_YESNOCANCEL       : cButtons = 3; break;
        case MB_YESNO             : cButtons = 2; break;
        case MB_RETRYCANCEL       : cButtons = 2; break;
        case MB_CANCELTRYCONTINUE : cButtons = 3; break;
    }

    int cMinButtons = (m_dwMmbFlags & MMB_Mask_MinNumButtons) / MMB_min1Button;
    if( cButtons < cMinButtons ) cButtons = cMinButtons;

    m_Controls.decreaseNumButtonsTo( cButtons,
                                     (m_uType & MB_HELP) != 0,
                                     !m_strLongMsg.IsEmpty() && !(m_dwMmbFlags & MMB_NoMoreButton) );

    if( m_Controls.m_hwndHelpBtn    ) SetWindowText( m_Controls.m_hwndHelpBtn,    MMBLanguageTable::getString(mmbid_Help, m_Language) );
    if( m_Controls.m_hwndMore       ) SetWindowText( m_Controls.m_hwndMore,       MMBLanguageTable::getString(mmbid_More__, m_Language) );
    if( m_Controls.m_ahwndChecks[0] ) SetWindowText( m_Controls.m_ahwndChecks[0], MMBLanguageTable::getString(mmbid_AlwaysShowThisMessage, m_Language) );

    switch( m_uMsgButtons )
    {
        default :
        case MB_OK :
            SetWindowLong( m_Controls.m_ahwndBtns[0], GWL_ID, IDOK );
            SetWindowText( m_Controls.m_ahwndBtns[0], MMBLanguageTable::getString(mmbid_Ok, m_Language) );
            break;
        case MB_OKCANCEL :
            SetWindowLong( m_Controls.m_ahwndBtns[0], GWL_ID, IDOK );
            SetWindowLong( m_Controls.m_ahwndBtns[1], GWL_ID, IDCANCEL );
            SetWindowText( m_Controls.m_ahwndBtns[0], MMBLanguageTable::getString(mmbid_Ok,     m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[1], MMBLanguageTable::getString(mmbid_Cancel, m_Language) );
            break;
        case MB_ABORTRETRYIGNORE :
            SetWindowLong( m_Controls.m_ahwndBtns[0], GWL_ID, IDABORT );
            SetWindowLong( m_Controls.m_ahwndBtns[1], GWL_ID, IDRETRY );
            SetWindowLong( m_Controls.m_ahwndBtns[2], GWL_ID, IDIGNORE );
            SetWindowText( m_Controls.m_ahwndBtns[0], MMBLanguageTable::getString(mmbid_Abort,  m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[1], MMBLanguageTable::getString(mmbid_Retry,  m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[2], MMBLanguageTable::getString(mmbid_Ignore, m_Language) );
            break;
        case MB_YESNOCANCEL :
            SetWindowLong( m_Controls.m_ahwndBtns[0], GWL_ID, IDYES );
            SetWindowLong( m_Controls.m_ahwndBtns[1], GWL_ID, IDNO );
            SetWindowLong( m_Controls.m_ahwndBtns[2], GWL_ID, IDCANCEL );
            SetWindowText( m_Controls.m_ahwndBtns[0], MMBLanguageTable::getString(mmbid_Yes,    m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[1], MMBLanguageTable::getString(mmbid_No,     m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[2], MMBLanguageTable::getString(mmbid_Cancel, m_Language) );
            break;
        case MB_YESNO :
            SetWindowLong( m_Controls.m_ahwndBtns[0], GWL_ID, IDYES );
            SetWindowLong( m_Controls.m_ahwndBtns[1], GWL_ID, IDNO );
            SetWindowText( m_Controls.m_ahwndBtns[0], MMBLanguageTable::getString(mmbid_Yes, m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[1], MMBLanguageTable::getString(mmbid_No,  m_Language) );
            break;
        case MB_RETRYCANCEL :
            SetWindowLong( m_Controls.m_ahwndBtns[0], GWL_ID, IDRETRY );
            SetWindowLong( m_Controls.m_ahwndBtns[1], GWL_ID, IDCANCEL );
            SetWindowText( m_Controls.m_ahwndBtns[0], MMBLanguageTable::getString(mmbid_Retry,  m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[1], MMBLanguageTable::getString(mmbid_Cancel, m_Language) );
            break;
        case MB_CANCELTRYCONTINUE :
            SetWindowLong( m_Controls.m_ahwndBtns[0], GWL_ID, IDCANCEL );
            SetWindowLong( m_Controls.m_ahwndBtns[1], GWL_ID, IDTRYAGAIN );
            SetWindowLong( m_Controls.m_ahwndBtns[2], GWL_ID, IDCONTINUE );
            SetWindowText( m_Controls.m_ahwndBtns[0], MMBLanguageTable::getString(mmbid_Cancel,   m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[1], MMBLanguageTable::getString(mmbid_TryAgain, m_Language) );
            SetWindowText( m_Controls.m_ahwndBtns[2], MMBLanguageTable::getString(mmbid_Continue, m_Language) );
            break;
    }

    if( m_Controls.m_ahwndBtns[0] && !m_strOwnBtn1.IsEmpty() ) SetWindowText( m_Controls.m_ahwndBtns[0], m_strOwnBtn1 );
    if( m_Controls.m_ahwndBtns[1] && !m_strOwnBtn2.IsEmpty() ) SetWindowText( m_Controls.m_ahwndBtns[1], m_strOwnBtn2 );
    if( m_Controls.m_ahwndBtns[2] && !m_strOwnBtn3.IsEmpty() ) SetWindowText( m_Controls.m_ahwndBtns[2], m_strOwnBtn3 );
    if( m_Controls.m_ahwndBtns[3] && !m_strOwnBtn4.IsEmpty() ) SetWindowText( m_Controls.m_ahwndBtns[3], m_strOwnBtn4 );
    if( m_Controls.m_hwndHelpBtn && !m_strOwnHelpBtn.IsEmpty() ) SetWindowText( m_Controls.m_hwndHelpBtn, m_strOwnHelpBtn );
    if( m_Controls.m_hwndMore && !m_strOwnMoreBtn.IsEmpty() ) SetWindowText( m_Controls.m_hwndMore, m_strOwnMoreBtn );
    if( m_Controls.m_ahwndChecks[0] && !m_strOwnCheck1.IsEmpty() ) SetWindowText( m_Controls.m_ahwndChecks[0], m_strOwnCheck1 );
    if( m_Controls.m_ahwndChecks[1] && !m_strOwnCheck2.IsEmpty() ) SetWindowText( m_Controls.m_ahwndChecks[1], m_strOwnCheck2 );

    // Now we evaluate different text message rects.
    RECT rectMsg_help = m_Controls.m_rectMsg;
    rectMsg_help.right = m_Controls.m_ptControlsAreaTopLeft.x + m_Controls.getCxMinControlsArea();
    if( !m_strMsg.IsEmpty() && m_Controls.m_hwndMsg )
    {
        // Now we evaluate the necessary height for the normal message text
        // and also possibly adjust its width.
        HDC hdc = GetDC( m_Controls.m_hwndMsg );
        HFONT hfont = (HFONT) SendMessage( m_Controls.m_hwndMsg, WM_GETFONT, 0, 0 );
        HGDIOBJ old_font = SelectObject( hdc, hfont );

        DrawText( hdc, m_strMsg, -1, &rectMsg_help, DT_CALCRECT | m_DT_MMB );

        SelectObject( hdc, old_font );
        ReleaseDC( m_Controls.m_hwndMsg, hdc );
    }

    rectMsg_help.right = m_Controls.correctControlsAreaNewRightCoord( rectMsg_help.right );

    RECT rectHeaderMsg_help = m_Controls.m_rectHeaderMsg;
    rectHeaderMsg_help.right = rectMsg_help.right;
    if( !m_strHeaderMsg.IsEmpty() && m_Controls.m_hwndHeaderMsg )
    {
        // Now we evaluate the necessary height for the Header message text.
        HDC hdc = GetDC( m_Controls.m_hwndHeaderMsg );
        HFONT hfont = (HFONT) SendMessage( m_Controls.m_hwndHeaderMsg, WM_GETFONT, 0, 0 );
        HGDIOBJ old_font = SelectObject( hdc, hfont );

        DrawText( hdc, m_strHeaderMsg, -1, &rectHeaderMsg_help, DT_CALCRECT | m_DT_MMBExtra );

        SelectObject( hdc, old_font );
        ReleaseDC( m_Controls.m_hwndHeaderMsg, hdc );
    }

    RECT rectExtraMsg_help = m_Controls.m_rectExtraMsg;
    rectExtraMsg_help.right = rectMsg_help.right;
    if( !m_strExtraMsg.IsEmpty() && m_Controls.m_hwndExtraMsg )
    {
        // Now we evaluate the necessary height for the Extra message text.
        HDC hdc = GetDC( m_Controls.m_hwndExtraMsg );
        HFONT hfont = (HFONT) SendMessage( m_Controls.m_hwndExtraMsg, WM_GETFONT, 0, 0 );
        HGDIOBJ old_font = SelectObject( hdc, hfont );

        DrawText( hdc, m_strExtraMsg, -1, &rectExtraMsg_help, DT_CALCRECT | m_DT_MMBExtra );

        SelectObject( hdc, old_font );
        ReleaseDC( m_Controls.m_hwndExtraMsg, hdc );
    }

    RECT rectExtraLink_help = m_Controls.m_rectExtraLink;
    rectExtraLink_help.right = rectMsg_help.right;
    if( !m_strExtraLink_Text.IsEmpty() && m_Controls.m_hwndExtraLink )
    {
        // Now we evaluate the necessary height for the Extra link text.
        HDC hdc = GetDC( m_Controls.m_hwndExtraLink );
        HFONT hfont = (HFONT) SendMessage( m_Controls.m_hwndExtraLink, WM_GETFONT, 0, 0 );
        HGDIOBJ old_font = SelectObject( hdc, hfont );

        DrawText( hdc, m_strExtraLink_Text, -1, &rectExtraLink_help, DT_CALCRECT | m_DT_MMBExtra );

        SelectObject( hdc, old_font );
        ReleaseDC( m_Controls.m_hwndExtraLink, hdc );
    }

    // Now we subclass all the message static controls.
    if( m_Controls.m_hwndHeaderMsg )
        SetWindowLongPtr( m_Controls.m_hwndHeaderMsg, GWLP_USERDATA, 
            SetWindowLongPtr( m_Controls.m_hwndHeaderMsg, GWLP_WNDPROC, (LONG_PTR) StaticCtrlProc ) );
    if( m_Controls.m_hwndMsg )
        SetWindowLongPtr( m_Controls.m_hwndMsg, GWLP_USERDATA, 
            SetWindowLongPtr( m_Controls.m_hwndMsg, GWLP_WNDPROC, (LONG_PTR) StaticCtrlProc ) );
    if( m_Controls.m_hwndExtraMsg )
        SetWindowLongPtr( m_Controls.m_hwndExtraMsg, GWLP_USERDATA, 
            SetWindowLongPtr( m_Controls.m_hwndExtraMsg, GWLP_WNDPROC, (LONG_PTR) StaticCtrlProc ) );
    if( m_Controls.m_hwndExtraLink )
        SetWindowLongPtr( m_Controls.m_hwndExtraLink, GWLP_USERDATA, 
            SetWindowLongPtr( m_Controls.m_hwndExtraLink, GWLP_WNDPROC, (LONG_PTR) UrlStaticCtrlProc ) );

    // Now we subclass the More button.
    if( m_Controls.m_hwndMore )
        SetWindowLongPtr( m_Controls.m_hwndMore, GWLP_USERDATA, 
            SetWindowLongPtr( m_Controls.m_hwndMore, GWLP_WNDPROC, (LONG_PTR) ButtonProc ) );

    m_Controls.adjustWidth( rectMsg_help.right );
    m_Controls.adjustHeightFromMsgs( rectHeaderMsg_help.bottom - rectHeaderMsg_help.top,
                                     rectMsg_help.bottom - rectMsg_help.top,
                                     rectExtraMsg_help.bottom - rectExtraMsg_help.top,
                                     rectExtraLink_help.bottom - rectExtraLink_help.top );

    int cCheckboxes = (m_dwMmbFlags & MMB_Mask_NumCheckboxes) / MMB_1Checkbox;
    m_Controls.decreaseNumCheckboxesTo( cCheckboxes );

    m_Controls.inflateBorders( m_rectInflateBorders );

    m_Controls.Collapse();
    m_Controls.setRects( false );
    Center( false, m_Controls.m_hwndMore ?
        (m_Controls.m_rectLongMsg.bottom - m_Controls.m_rectLongMsg.top) * 2 / 3 : 0 );
    // Now we ensure that the entire dialog fits within the desktop area.
    // The dialog will be eventually repositioned if necessary.
    SendMessage( m_hWnd, DM_REPOSITION, 0, 0 );

    SetWindowText( m_hWnd, m_strCaption );
    if( m_Controls.m_hwndHeaderMsg ) SetWindowText( m_Controls.m_hwndHeaderMsg, m_strHeaderMsg );
    if( m_Controls.m_hwndMsg ) SetWindowText( m_Controls.m_hwndMsg, m_strMsg );
    if( m_Controls.m_hwndLongMsg ) SetWindowText( m_Controls.m_hwndLongMsg, m_strLongMsg );

    if( m_Controls.m_hwndExtraMsg ) SetWindowText( m_Controls.m_hwndExtraMsg, m_strExtraMsg );
    if( m_Controls.m_hwndExtraLink ) SetWindowText( m_Controls.m_hwndExtraLink, m_strExtraLink_Text );

    if( m_Controls.m_ahwndChecks[0] )
        SendMessage( m_Controls.m_ahwndChecks[0], BM_SETCHECK, m_bIs_Check1_Checked ? BST_CHECKED : BST_UNCHECKED, 0 );
    if( m_Controls.m_ahwndChecks[1] )
        SendMessage( m_Controls.m_ahwndChecks[1], BM_SETCHECK, m_bIs_Check2_Checked ? BST_CHECKED : BST_UNCHECKED, 0 );

    if( m_uMsgBeep ) MessageBeep( m_uMsgBeep );
    if( m_dwMmbFlags & MMB_PushForeground ) PushForegroundWindow( m_hWnd );
    else if( m_uType & MB_SETFOREGROUND ) SetForegroundWindow( m_hWnd );

    switch( m_uMsgDefButton )
    {
        case MB_DEFBUTTON1 : m_hwndDefBtn = m_Controls.m_ahwndBtns[0]; break;
        case MB_DEFBUTTON2 : m_hwndDefBtn = m_Controls.m_ahwndBtns[1]; break;
        case MB_DEFBUTTON3 : m_hwndDefBtn = m_Controls.m_ahwndBtns[2]; break;
        case MB_DEFBUTTON4 : m_hwndDefBtn = m_Controls.m_ahwndBtns[3]; break;
        default            : m_hwndDefBtn = NULL; break;
    }

    if( !m_hwndDefBtn )
        return TRUE;

    SendMessage( m_hWnd, DM_SETDEFID, (WPARAM) GetDlgCtrlID( m_hwndDefBtn ), 0 );
    SendMessage( m_hWnd, WM_NEXTDLGCTL, (WPARAM) m_hwndDefBtn, TRUE );

    return FALSE;
}

void MoreMessageBox::onDestroy()
{
    if( m_hfontBold ) DeleteObject( m_hfontBold );
    if( m_hfontUnderlined ) DeleteObject( m_hfontUnderlined );
}

///<summary>
///  <b>onMoreClicked()</b> is called when the More button has been clicked.
///  It is used to expand or collapse the MoreMessageBox.
///</summary>
void MoreMessageBox::onMoreClicked()
{
    HWND hwndSetFocus = m_hwndDefBtn;
    if( !hwndSetFocus ) hwndSetFocus = m_Controls.m_ahwndBtns[0];
    if( !hwndSetFocus ) hwndSetFocus = m_Controls.m_hwndMore;
    if( hwndSetFocus )
        SendMessage( m_hWnd, WM_NEXTDLGCTL, (WPARAM) hwndSetFocus, TRUE );

    if( !m_Controls.m_bIsExpanded )
        m_Controls.Expand();
    else
        m_Controls.Collapse();

    m_Controls.setRects();
}

void MoreMessageBox::onUrlClicked()
{
    if( m_idMessage_UrlClicked )
    {
        SendMessage( GetParent(m_hWnd), m_idMessage_UrlClicked, 0, (LPARAM) m_hWnd );
        return;
    }

    if( !m_strExtraLink_Url.IsEmpty() )
        ShellExecute( NULL, NULL, m_strExtraLink_Url, NULL, NULL, SW_SHOWNORMAL );
}

INT_PTR CALLBACK MoreMessageBox::WndProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    switch( message )
    {
        case WM_INITDIALOG :
        {
            MoreMessageBox *pBox = (MoreMessageBox*) lParam;
            pBox->m_hWnd = hwnd;
            SetWindowLongPtr( hwnd, GWLP_USERDATA, (LONG_PTR) pBox );

            return pBox->onInitDialog();
        }

        case WM_COMMAND :
        {
            MoreMessageBox *pBox = (MoreMessageBox*) GetWindowLongPtr( hwnd, GWLP_USERDATA );
            if( !pBox )
                { EndDialog( hwnd, IDCANCEL ); return TRUE; }

            if( (pBox->m_Controls.m_ahwndBtns[0] && pBox->m_Controls.m_ahwndBtns[0] == (HWND) lParam) ||
                (pBox->m_Controls.m_ahwndBtns[1] && pBox->m_Controls.m_ahwndBtns[1] == (HWND) lParam) ||
                (pBox->m_Controls.m_ahwndBtns[2] && pBox->m_Controls.m_ahwndBtns[2] == (HWND) lParam) ||
                (pBox->m_Controls.m_ahwndBtns[3] && pBox->m_Controls.m_ahwndBtns[3] == (HWND) lParam) )
            {
                EndDialog( hwnd, LOWORD(wParam) );
                return TRUE;
            }

            if( LOWORD(wParam) == IDCANCEL )
            {
                if( GetDlgItem(hwnd, IDCANCEL) ) { EndDialog( hwnd, IDCANCEL ); return TRUE; }
                if( GetDlgItem(hwnd, IDOK    ) ) { EndDialog( hwnd, IDOK     ); return TRUE; }
            }

            switch( LOWORD(wParam) )
            {
                case mmbIDC_HELPBTN :
                {
                    // We just forward the WM_HELP execution, so we leave the struct empty.

                    HELPINFO hi;
                    memset( &hi, 0, sizeof(hi) );

                    SendMessage( hwnd, WM_HELP, 0, (LPARAM) &hi );

                    return TRUE;
                }

                case mmbIDC_MORE :
                {
                    pBox->onMoreClicked();
                    return TRUE;
                }

                case mmbIDC_CHECK1 :
                {
                    pBox->m_bIs_Check1_Checked =
                        SendMessage( pBox->m_Controls.m_ahwndChecks[0], BM_GETCHECK, 0, 0 ) == BST_CHECKED;
                    return TRUE;
                }

                case mmbIDC_CHECK2 :
                {
                    pBox->m_bIs_Check2_Checked =
                        SendMessage( pBox->m_Controls.m_ahwndChecks[1], BM_GETCHECK, 0, 0 ) == BST_CHECKED;
                    return TRUE;
                }
            }

            break;
        }

        case WM_MMB_URLCLICKED :
        {
            MoreMessageBox *pBox = (MoreMessageBox*) GetWindowLongPtr( hwnd, GWLP_USERDATA );
            if( !pBox ) break;

            pBox->onUrlClicked();

            return TRUE;
        }

        case WM_HELP :      // is invoked automatically for us when the user presses the F1 key
        {
            MoreMessageBox *pBox = (MoreMessageBox*) GetWindowLongPtr( hwnd, GWLP_USERDATA );
            if( !pBox ) break;

            if( !pBox->m_Controls.m_hwndHelpBtn )
                break;

            HWND hwndParent = GetParent( hwnd );
            if( !hwndParent ) break;

            HELPINFO hi;
            memset( &hi, 0, sizeof(hi) );
            hi.cbSize = sizeof(hi);
            hi.iContextType = HELPINFO_WINDOW;
            hi.iCtrlId = IDHELP;
            hi.hItemHandle = hwnd;
            hi.dwContextId = 0;
            GetCursorPos( &hi.MousePos );

            SendMessage( hwndParent, WM_HELP, 0, (LPARAM) &hi );

            return TRUE;
        }

        case WM_MMB_DRAWSTATIC :
        case WM_MMB_CALCSTATIC :
        {
            MoreMessageBox *pBox = (MoreMessageBox*) GetWindowLongPtr( hwnd, GWLP_USERDATA );
            if( !pBox ) break;

            bool bDrawing = message == WM_MMB_DRAWSTATIC;

            PRECT pRect    = (PRECT) wParam;    // valid only when calculating
            HDC   hdc      = (HDC)   wParam;    // valid only when drawing
            HWND  hwndCtrl = (HWND)  lParam;

            RECT rect;
            GetClientRect( hwndCtrl, &rect );

            if( bDrawing )
            {
                HBRUSH hbr = (HBRUSH) SendMessage( hwnd, WM_CTLCOLORSTATIC, (WPARAM) hdc, (LPARAM) hwnd );
                FillRect( hdc, &rect, hbr );
            }

            String s;
            s.GetWindowtext( hwndCtrl );

            if( !bDrawing )
                hdc = GetDC( hwndCtrl );

            HFONT hfont = (HFONT) SendMessage( hwndCtrl, WM_GETFONT, 0, 0 );
            HGDIOBJ old_font = SelectObject( hdc, hfont );
            int old_bkmd = SetBkMode( hdc, TRANSPARENT );

            UINT fmt = hwndCtrl == pBox->m_Controls.m_hwndMsg ? pBox->m_DT_MMB :
                       hwndCtrl == pBox->m_Controls.m_hwndHeaderMsg ? pBox->m_DT_MMBHeader :
                       pBox->m_DT_MMBExtra;
            if( !bDrawing ) fmt |= DT_CALCRECT;

            DrawText( hdc, s, -1, &rect, fmt );

            SetBkMode( hdc, old_bkmd );
            SelectObject( hdc, old_font );

            if( !bDrawing )
                ReleaseDC( hwndCtrl, hdc );

            if( !bDrawing && pRect )
                *pRect = rect;

            return TRUE;
        }

        case WM_MMB_DRAWBUTTON :
        {
            MoreMessageBox *pBox = (MoreMessageBox*) GetWindowLongPtr( hwnd, GWLP_USERDATA );
            if( !pBox ) break;

            HDC  hdc      = (HDC)  wParam;
            HWND hwndCtrl = (HWND) lParam;

            RECT rect;
            GetClientRect( hwndCtrl, &rect );

            RECT rectUpDownArrow = rect;
            rectUpDownArrow.right -= pBox->m_nOffsetUpDownArrow;
            rectUpDownArrow.left = rectUpDownArrow.right - pBox->m_sizeUpDownArrow.cx;
            rectUpDownArrow.bottom = rect.bottom - ((rect.bottom - rect.top) - pBox->m_sizeUpDownArrow.cy) / 2;
            rectUpDownArrow.top = rectUpDownArrow.bottom - pBox->m_sizeUpDownArrow.cy;

            if( pBox->m_hiconUpArrow &&
                pBox->m_Controls.m_bIsExpanded &&
                !(pBox->m_dwMmbFlags & MMB_ExpandableOnly) )
            {
                DrawState( hdc, GetSysColorBrush(COLOR_GRAYTEXT), NULL,
                           (LPARAM) pBox->m_hiconUpArrow, 0,
                           rectUpDownArrow.left, rectUpDownArrow.top,
                           pBox->m_sizeUpDownArrow.cx, pBox->m_sizeUpDownArrow.cy,
                           DST_ICON | (IsWindowEnabled(hwndCtrl) ? DSS_NORMAL : DSS_MONO) );
            }
            else
            if( pBox->m_hiconDownArrow )
            {
                //DrawIconEx( hdc, rectUpDownArrow.left, rectUpDownArrow.top, pBox->m_hiconDownArrow,
                //            pBox->m_sizeDownArrow.cx, pBox->m_sizeDownArrow.cy, 0, NULL, DI_NORMAL );

                DrawState( hdc, GetSysColorBrush(COLOR_GRAYTEXT), NULL,
                           (LPARAM) pBox->m_hiconDownArrow, 0,
                           rectUpDownArrow.left, rectUpDownArrow.top,
                           pBox->m_sizeUpDownArrow.cx, pBox->m_sizeUpDownArrow.cy,
                           DST_ICON | (IsWindowEnabled(hwndCtrl) ? DSS_NORMAL : DSS_MONO) );
            }

            return TRUE;
        }

        case WM_DESTROY :
        {
            MoreMessageBox *pBox = (MoreMessageBox*) GetWindowLongPtr( hwnd, GWLP_USERDATA );
            if( !pBox ) break;

            pBox->onDestroy();

            break;
        }
    }

    return FALSE;
}

///<summary>
///  <b>StaticCtrlProc()</b> is the procedure for our subclassed static control
///  used to display the normal message text. Since a normal static control
///  does not break a line that is too long and does not contain space chars,
///  we must subclass the control and ensure the line breaks on ourselves. <br/>
///  This can be realized specifically using the DT_EDITCONTROL flag in DrawText()
///  which is why we subclassed the control.
///</summary>
///<remarks>
///  This proc expects the GWLP_USERDATA of the control to contain the control's
///  original window proc.
///</remarks>
LRESULT CALLBACK MoreMessageBox::StaticCtrlProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    #ifdef STRICT
    WNDPROC oldproc = (WNDPROC) GetWindowLongPtr( hwnd, GWLP_USERDATA );
    #else
    FARPROC oldproc = (FARPROC) GetWindowLongPtr( hwnd, GWLP_USERDATA );
    #endif
    if( !oldproc ) return DefWindowProc( hwnd, message, wParam, lParam );

    switch( message )
    {
        case WM_PRINTCLIENT :
        {
            if( (lParam & PRF_CLIENT) )
            {
                HWND hwndParent = GetParent( hwnd );
                SendMessage( hwndParent, WM_MMB_DRAWSTATIC, wParam, (LPARAM) hwnd );
            }

            return 0;
        }

        case WM_PAINT :
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint( hwnd, &ps );

            RECT rect;
            // Note: The following #GetClipBox()# will return NULLREGION if the
            // DC/window repaint area is completely obsured by other windows.
            // So in that case, we do not need to draw anything.
            if( GetClipBox(hdc, &rect) != NULLREGION )
                SendMessage( hwnd, WM_PRINT, (WPARAM) hdc, PRF_CHECKVISIBLE | PRF_CHILDREN | PRF_CLIENT | PRF_ERASEBKGND | PRF_NONCLIENT );

            EndPaint( hwnd, &ps );
            return 0;
        }
    }

    return CallWindowProc( oldproc, hwnd, message, wParam, lParam );
}

LRESULT CALLBACK MoreMessageBox::UrlStaticCtrlProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    switch( message )
    {
        case WM_LBUTTONDOWN :
        {
            RECT rect;

            HWND hwndParent = GetParent( hwnd );
            SendMessage( hwndParent, WM_MMB_CALCSTATIC, (WPARAM) &rect, (LPARAM) hwnd );

            POINT mouse;
            mouse.x = (short) LOWORD(lParam);
            mouse.y = (short) HIWORD(lParam);

            if( PtInRect(&rect, mouse) )
                SendMessage( hwndParent, WM_MMB_URLCLICKED, 0, (LPARAM) hwnd );

            break;
        }

        case WM_SETCURSOR :
        {
            DWORD dw = GetMessagePos();

            RECT rect;

            HWND hwndParent = GetParent( hwnd );
            SendMessage( hwndParent, WM_MMB_CALCSTATIC, (WPARAM) &rect, (LPARAM) hwnd );

            POINT mouse;
            mouse.x = (short) LOWORD(dw);
            mouse.y = (short) HIWORD(dw);
            ScreenToClient( hwnd, &mouse );

            if( PtInRect(&rect, mouse) )
            {
                HCURSOR hcrsrHand = LoadCursor( NULL, IDC_HAND );
                if( hcrsrHand )
                    { SetCursor( hcrsrHand ); return TRUE; }
            }

            break;
        }
    }

    return StaticCtrlProc( hwnd, message, wParam, lParam );
}

///<summary>
///  <b>ButtonProc()</b> is the procedure for our subclassed button controls
///  used to display also an icon besides the text.
///</summary>
///<remarks>
///  This proc expects the GWLP_USERDATA of the control to contain the control's
///  original window proc.
///</remarks>
LRESULT CALLBACK MoreMessageBox::ButtonProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    #ifdef STRICT
    WNDPROC oldproc = (WNDPROC) GetWindowLongPtr( hwnd, GWLP_USERDATA );
    #else
    FARPROC oldproc = (FARPROC) GetWindowLongPtr( hwnd, GWLP_USERDATA );
    #endif
    if( !oldproc ) return DefWindowProc( hwnd, message, wParam, lParam );

    switch( message )
    {
        case WM_PRINTCLIENT :
        {
            LRESULT lr = CallWindowProc( oldproc, hwnd, message, wParam, lParam );

            if( (lParam & PRF_CLIENT) )
            {
                HWND hwndParent = GetParent( hwnd );
                SendMessage( hwndParent, WM_MMB_DRAWBUTTON, wParam, (LPARAM) hwnd );
            }

            return lr;
        }

        case WM_PAINT :
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint( hwnd, &ps );

            RECT rect;
            // Note: The following #GetClipBox()# will return NULLREGION if the
            // DC/window repaint area is completely obsured by other windows.
            // So in that case, we do not need to draw anything.
            if( GetClipBox(hdc, &rect) != NULLREGION )
                SendMessage( hwnd, WM_PRINT, (WPARAM) hdc, PRF_CHECKVISIBLE | PRF_CHILDREN | PRF_CLIENT | PRF_ERASEBKGND | PRF_NONCLIENT );

            EndPaint( hwnd, &ps );
            return 0;
        }
    }

    return CallWindowProc( oldproc, hwnd, message, wParam, lParam );
}

void MoreMessageBox::EnableDisabledPopups( bool bEnable )
{
    HWND    *ptr = m_aDisabledPopups.Ptr();
    _size_t_ cnt = m_aDisabledPopups.GetSize();
    if( !ptr || cnt <= 0 ) return;

    for( _idx_ i = 0 ; i < (_idx_) cnt ; ++i, ++ptr )
        if( *ptr )
            EnableWindow( *ptr, bEnable );
}

BOOL CALLBACK MoreMessageBox::FindWndProc( HWND hwnd, LPARAM lParam )
{
    if( !IsWindowVisible(hwnd) )
        return TRUE;

    DWORD ctId = GetCurrentThreadId();
    HWND *phwndFound = (HWND*) lParam;

    // Priorities:
    //  0   no window
    //  1   top-level-window within another thread.
    //  2   non-active top-level-window within the current thread.
    //  3   the active top-level-window within the current thread.

    const int priorityMax = 3;

    int priorityWndFound =
        (*phwndFound == NULL) ? 0 :
        (GetWindowThreadProcessId(*phwndFound, NULL) != ctId) ? 1 :
        (GetActiveWindow() != *phwndFound) ? 2 : 3;

    // Now we get the top-level-window
    hwnd = GetAncestor( hwnd, GA_ROOTOWNER );
    if( hwnd == *phwndFound ) return TRUE;

    int priorityWnd =
        (hwnd == NULL) ? 0 :
        (GetWindowThreadProcessId(hwnd, NULL) != ctId) ? 1 :
        (GetActiveWindow() != hwnd) ? 2 : 3;

    if( priorityWnd > priorityWndFound )
        { *phwndFound = hwnd; priorityWndFound = priorityWnd; }

    return priorityWndFound < priorityMax;
}

///<summary>
///  <b>FindNiceParent()</b> searches the windows of the current thread
///  and returns the first top-level-window of the current thread.
///  If this is not found, it returns any top-level-window of one of
///  the current thread's windows. Otherwise it returns NULL. <br/>
///  Note: But invisible windows will be skipped. <br/>
///  Note: If there are no windows created within the current thread,
///  NULL will be returned.
///</summary>
HWND MoreMessageBox::FindNiceParent()
{
    HWND hwndFound = NULL;

    EnumThreadWindows( GetCurrentThreadId(), FindWndProc, (LPARAM) &hwndFound );

    return hwndFound;
}

///<summary>
///  <b>PushForegroundWindow()</b> acts similar to <b>SetForegroundWindow()</b>
///  except for the difference, that even on Win2k/XP it will bring the window
///  really to the foreground, preventing the taskbar from flashing.
///</summary>
///<remarks>
///  This call will *not* cause any system parameters to be changed as mentioned
///  in the MSDN to call <pre>
///     SystemParametersInfo( SPI_SETFOREGROUNDLOCKTIMEOUT, 0, NULL, SPIF_SENDWININICHANGE | SPIF_UPDATEINIFILE );
///  </pre> to change the foreground window behaviour for all windows
///  in the current system.
///  <br/>
///  Instead, it will attach temporarily to the thread of the foreground process's thread.
///</remarks>
void MoreMessageBox::PushForegroundWindow( HWND hwnd )
{
    DWORD idthrdCurrent    = GetCurrentThreadId();
    DWORD idthrdForeground = GetWindowThreadProcessId( ::GetForegroundWindow(), NULL );

    if( idthrdCurrent != idthrdForeground )
        AttachThreadInput( idthrdForeground, idthrdCurrent, TRUE );

    SetForegroundWindow( hwnd );

    if( idthrdCurrent != idthrdForeground )
        AttachThreadInput( idthrdForeground, idthrdCurrent, FALSE );
}

#endif

void MoreMessageBox::CenterWindowOverWindow( HWND hwndToBeCentered, HWND hwndReferenceWindow, bool bCenterOnlyHorizontally, long nAdditionalHeight )
{
    if( !hwndToBeCentered ) return;

    RECT rectRef;
    if( hwndReferenceWindow ) GetWindowRect( hwndReferenceWindow, &rectRef );
    else SystemParametersInfo( SPI_GETWORKAREA, 0, &rectRef, 0 );

    RECT rect;
    GetWindowRect( hwndToBeCentered, &rect );

    long dx = ((rectRef.right - rectRef.left) - (rect.right - rect.left)) / 2 + (rectRef.left - rect.left);
    long dy = ((rectRef.bottom - rectRef.top) - (rect.bottom - rect.top + nAdditionalHeight)) / 2 + (rectRef.top - rect.top );
    if( bCenterOnlyHorizontally ) dy = 0;
    OffsetRect( &rect, dx, dy );

    SetWindowPos( hwndToBeCentered, HWND_TOP,
                  rect.left, rect.top,
                  rect.right - rect.left, rect.bottom - rect.top,
                  SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
}

void MoreMessageBox::CenterWindowOverItsParent( HWND hwndToBeCentered, bool bCenterOnlyHorizontally, long nAdditionalHeight )
{
    CenterWindowOverWindow( hwndToBeCentered, GetParent(hwndToBeCentered), bCenterOnlyHorizontally, nAdditionalHeight );
}

HFONT MoreMessageBox::CreateAdoptedFont( HFONT hfontFrom, int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut )
{
    if( !hfontFrom ) return NULL;

    LOGFONT lf;
    memset( &lf, 0, sizeof(lf) );
    if( ::GetObject( hfontFrom, sizeof(lf), &lf ) == 0 )
        return NULL;

    lf.lfWeight    = nWeight;
    lf.lfItalic    = bItalic;
    lf.lfUnderline = bUnderline;
    lf.lfStrikeOut = bStrikeOut;

    return CreateFontIndirect( &lf );
}

#ifndef _WIN32_WCE

void MoreMessageBox::Controls::Empty()
{
    memset( this, 0, sizeof(MoreMessageBox::Controls) );
}

void MoreMessageBox::Controls::getFromDialog( HWND hwnd,
                                              DWORD dwMmbFlags, int minWidthAsNumButtons )
{
    if( minWidthAsNumButtons < 1 ) minWidthAsNumButtons = 1;
    if( minWidthAsNumButtons > MaxButtons ) minWidthAsNumButtons = MaxButtons;

    m_dwMmbFlags = dwMmbFlags;
    m_minWidthAsNumButtons = minWidthAsNumButtons;

    m_hWnd = hwnd;

    m_hwndHeaderMsg  = GetDlgItem( m_hWnd, mmbIDC_HEADER );
    m_hwndMsgIcon    = GetDlgItem( m_hWnd, mmbIDC_MSGICON );
    m_hwndMsg        = GetDlgItem( m_hWnd, mmbIDC_MSG );
    m_hwndLongMsg    = GetDlgItem( m_hWnd, mmbIDC_LONGMSG );
    m_ahwndBtns[0]   = GetDlgItem( m_hWnd, mmbIDC_BUTTON1 );
    m_ahwndBtns[1]   = GetDlgItem( m_hWnd, mmbIDC_BUTTON2 );
    m_ahwndBtns[2]   = GetDlgItem( m_hWnd, mmbIDC_BUTTON3 );
    m_ahwndBtns[3]   = GetDlgItem( m_hWnd, mmbIDC_BUTTON4 );
    m_hwndHelpBtn    = GetDlgItem( m_hWnd, mmbIDC_HELPBTN );
    m_hwndMore       = GetDlgItem( m_hWnd, mmbIDC_MORE );
    m_ahwndChecks[0] = GetDlgItem( m_hWnd, mmbIDC_CHECK1 );
    m_ahwndChecks[1] = GetDlgItem( m_hWnd, mmbIDC_CHECK2 );
    m_hwndHr         = GetDlgItem( m_hWnd, mmbIDC_HR );
    m_hwndExtraIcon  = GetDlgItem( m_hWnd, mmbIDC_EXTRAICON );
    m_hwndExtraMsg   = GetDlgItem( m_hWnd, mmbIDC_EXTRAMSG );
    m_hwndExtraLink  = GetDlgItem( m_hWnd, mmbIDC_EXTRALINK );

    GetWindowRect( m_hWnd, &m_rectWnd );
    GetWindowRect( m_hwndHeaderMsg, &m_rectHeaderMsg );
    GetWindowRect( m_hwndMsgIcon, &m_rectMsgIcon );
    GetWindowRect( m_hwndMsg, &m_rectMsg );
    GetWindowRect( m_hwndLongMsg, &m_rectLongMsg );
    GetWindowRect( m_ahwndBtns[0], &m_arectBtns[0] );
    GetWindowRect( m_ahwndBtns[1], &m_arectBtns[1] );
    GetWindowRect( m_ahwndBtns[2], &m_arectBtns[2] );
    GetWindowRect( m_ahwndBtns[3], &m_arectBtns[3] );
    GetWindowRect( m_hwndHelpBtn, &m_rectHelpBtn );
    GetWindowRect( m_hwndMore, &m_rectMore );
    GetWindowRect( m_ahwndChecks[0], &m_arectChecks[0] );
    GetWindowRect( m_ahwndChecks[1], &m_arectChecks[1] );
    GetWindowRect( m_hwndHr, &m_rectHr );
    GetWindowRect( m_hwndExtraIcon, &m_rectExtraIcon );
    GetWindowRect( m_hwndExtraMsg, &m_rectExtraMsg );
    GetWindowRect( m_hwndExtraLink, &m_rectExtraLink );

    long cxIcon = GetSystemMetrics( SM_CXICON );
    long cyIcon = GetSystemMetrics( SM_CYICON );
    m_rectMsgIcon.right  = m_rectMsgIcon.left + cxIcon;
    m_rectMsgIcon.bottom = m_rectMsgIcon.top  + cyIcon;
    m_rectExtraIcon.right  = m_rectExtraIcon.left + cxIcon;
    m_rectExtraIcon.bottom = m_rectExtraIcon.top  + cyIcon;

    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectHeaderMsg, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectMsgIcon, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectMsg, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectLongMsg, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_arectBtns[0], 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_arectBtns[1], 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_arectBtns[2], 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_arectBtns[3], 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectHelpBtn, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectMore, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_arectChecks[0], 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_arectChecks[1], 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectHr, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectExtraIcon, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectExtraMsg, 2 );
    MapWindowPoints( NULL, m_hWnd, (PPOINT) &m_rectExtraLink, 2 );

    RECT rectClient;
    GetClientRect( m_hWnd, &rectClient );

    m_bIsExpanded = true;
    m_dyExpand = m_arectBtns[0].top - m_rectLongMsg.top;
    m_dxButtons = m_arectBtns[1].left - m_arectBtns[0].left;
    m_dyCheck1 = m_arectChecks[0].bottom - m_rectMore.bottom;
    m_dyCheckboxes = m_arectChecks[1].top - m_arectChecks[0].top;
    m_rectControlsAreaBordersThicknesses.left   = m_arectBtns[0].left - rectClient.left;
    m_rectControlsAreaBordersThicknesses.right  = rectClient.right - m_rectMore.right;
    m_rectControlsAreaBordersThicknesses.top    = m_rectMsg.top - rectClient.top;
    m_rectControlsAreaBordersThicknesses.bottom = rectClient.bottom - m_arectChecks[1].bottom;
    m_ptControlsAreaTopLeft.x = m_arectBtns[0].left;
    m_ptControlsAreaTopLeft.y = m_rectMsg.top;
    m_cxminControlsArea = m_arectBtns[m_minWidthAsNumButtons - 1].right - m_arectBtns[0].left;   // minimum width = 1 to 4 buttons at all
    m_dyExtraLinkDist = m_rectExtraLink.top - m_rectExtraMsg.bottom;
}

void MoreMessageBox::Controls::deleteHeaderMsg()
{
    if( m_hwndHeaderMsg )
    {
        DestroyWindow( m_hwndHeaderMsg );
        m_hwndHeaderMsg = NULL;

        long dy = m_rectHeaderMsg.top - m_rectMsg.top;

        offsetControls( 0, dy, true, false, true, true, true, true, true, true, true );
    }
}

void MoreMessageBox::Controls::deleteMsgIcon()
{
    if( m_hwndMsgIcon )
    {
        DestroyWindow( m_hwndMsgIcon );
        m_hwndMsgIcon = NULL;

        // Since MsgIcon does not exist any more,
        // we stretch the normal and long message controls towards the left.
        m_rectMsg.left =
        m_rectLongMsg.left = m_ptControlsAreaTopLeft.x;
    }
}

void MoreMessageBox::Controls::getMsgIconSize()
{
    if( !m_hwndMsgIcon ) return;

    RECT rc;
    GetWindowRect( m_hwndMsgIcon, &rc );

    m_rectMsgIcon.right  = m_rectMsgIcon.left + (rc.right  - rc.left);
    m_rectMsgIcon.bottom = m_rectMsgIcon.top  + (rc.bottom - rc.top );
}

void MoreMessageBox::Controls::getExtraIconSize()
{
    if( !m_hwndExtraIcon ) return;

    RECT rc;
    GetWindowRect( m_hwndExtraIcon, &rc );

    m_rectExtraIcon.right  = m_rectExtraIcon.left + (rc.right  - rc.left);
    m_rectExtraIcon.bottom = m_rectExtraIcon.top  + (rc.bottom - rc.top );
}

///<remarks>
///  IMPORTANT: Do *not* call this *after* having called <b>decreaseNumCheckboxesTo()</b>
///  because this algo needs a still existing last checkbox.
///</remarks>
void MoreMessageBox::Controls::deleteExtraArea( bool bDelIcon, bool bDelMsg, bool bDelLink )
{
    if( !bDelIcon && !bDelMsg && !bDelLink ) return;
    if( !m_hwndExtraIcon && !m_hwndExtraMsg && !m_hwndExtraLink ) return;

    RECT rectOldExtra = { m_rectWnd.right - m_rectWnd.left, m_rectWnd.bottom - m_rectWnd.top, 0, 0 };
    if( m_hwndExtraMsg  ) rectOldExtra = m_rectExtraMsg;
    if( m_hwndExtraLink ) { RECT rTmp = rectOldExtra; UnionRect( &rectOldExtra, &rTmp, &m_rectExtraLink ); }
    if( m_hwndExtraIcon ) { RECT rTmp = rectOldExtra; UnionRect( &rectOldExtra, &rTmp, &m_rectExtraIcon ); }
    long cyOldExtra = rectOldExtra.bottom - rectOldExtra.top;

    if( m_hwndExtraIcon && bDelIcon )
    {
        DestroyWindow( m_hwndExtraIcon );
        m_hwndExtraIcon = NULL;

        // Since ExtraIcon does not exist any more,
        // we stretch the extra message and extra link controls towards the left.
        m_rectExtraMsg.left =
        m_rectExtraLink.left = m_ptControlsAreaTopLeft.x;
    }

    if( m_hwndExtraMsg && bDelMsg )
    {
        DestroyWindow( m_hwndExtraMsg );
        m_hwndExtraMsg = NULL;

        long dy = m_rectExtraMsg.top - m_rectExtraLink.top;
        OffsetRect( &m_rectExtraLink, 0, dy );
    }

    if( m_hwndExtraLink && bDelLink )
    {
        DestroyWindow( m_hwndExtraLink );
        m_hwndExtraLink = NULL;
    }

    // Let's check if the whole Extra area has disappeared and if
    // so, we delete the horz ruler above it and make the dialog shorter.
    if( !m_hwndExtraIcon && !m_hwndExtraMsg && !m_hwndExtraLink )
    {
        DestroyWindow( m_hwndHr );
        m_hwndHr = NULL;

        long dy = m_arectChecks[MaxCheckboxes - 1].bottom - rectOldExtra.bottom;
        m_rectWnd.bottom += dy;
        return;
    }

    RECT rectNewExtra = { m_rectWnd.right - m_rectWnd.left, m_rectWnd.bottom - m_rectWnd.top, 0, 0 };
    if( m_hwndExtraMsg  ) rectNewExtra = m_rectExtraMsg;
    if( m_hwndExtraLink ) { RECT rTmp = rectNewExtra; UnionRect( &rectNewExtra, &rTmp, &m_rectExtraLink ); }
    if( m_hwndExtraIcon ) { RECT rTmp = rectNewExtra; UnionRect( &rectNewExtra, &rTmp, &m_rectExtraIcon ); }
    long cyNewExtra = rectNewExtra.bottom - rectNewExtra.top;

    long dy = cyNewExtra - cyOldExtra;
    m_rectWnd.bottom += dy;
}

///<summary>
///  <b>decreaseNumButtonsTo()</b> decreases the number of buttons currently
///  existing, or leaves them as they are (but never can increase them again
///  because the unnecessary controls will be destroyed).
///</summary>
///<param name="cButtons"> The new number of buttons that should exist.
///  This number can only be less or equal to the current number of buttons. </param>
///<param name="bHelpBtn"> if the help button shall exist </param>
///<param name="bMoreBtn"> if the more button shall exist </param>
void MoreMessageBox::Controls::decreaseNumButtonsTo( int cButtons, bool bHelpBtn, bool bMoreBtn )
{
    for( int i = MaxButtons - 1 ; i >= 0 && cButtons <= i ; --i )
        if( m_ahwndBtns[i] )
        {
            m_rectMore = m_rectHelpBtn;
            m_rectHelpBtn = m_arectBtns[i];

            DestroyWindow( m_ahwndBtns[i] );
            m_ahwndBtns[i] = NULL;
        }

    if( !bHelpBtn && m_hwndHelpBtn )
    {
        m_rectMore = m_rectHelpBtn;

        DestroyWindow( m_hwndHelpBtn );
        m_hwndHelpBtn = NULL;
    }

    if( !bMoreBtn && m_hwndMore )
    {
        DestroyWindow( m_hwndMore );
        m_hwndMore = NULL;
    }
}

RECT MoreMessageBox::Controls::getAllButtonsBoundsRect()
{
    RECT rect = m_arectBtns[0];

    if( m_hwndMore )
        { rect.right = m_rectMore.right; return rect; }

    if( m_hwndHelpBtn )
        { rect.right = m_rectHelpBtn.right; return rect; }

    for( int i = MaxButtons - 1 ; i >= 0 ; --i )
        if( m_ahwndBtns[i] )
            { rect.right = m_arectBtns[i].right; return rect; }

    // No buttons -> empty rect
    rect.right = rect.left;
    return rect;
}

long MoreMessageBox::Controls::getCxMinControlsArea()
{
    RECT rectBtns = getAllButtonsBoundsRect();
    long cxmin = rectBtns.right - rectBtns.left;
    if( cxmin < m_cxminControlsArea ) cxmin = m_cxminControlsArea;
    return cxmin;
}

long MoreMessageBox::Controls::correctControlsAreaNewRightCoord( long nControlsAreaNewRightCoord )
{
    long cxmin = getCxMinControlsArea();

    if( nControlsAreaNewRightCoord < m_ptControlsAreaTopLeft.x + cxmin )
        nControlsAreaNewRightCoord = m_ptControlsAreaTopLeft.x + cxmin;

    return nControlsAreaNewRightCoord;
}

///<remarks> Call this only once or else the message controls could
///  get wrong widths. </remarks>
void MoreMessageBox::Controls::adjustWidth( long nControlsAreaNewRightCoord )
{
    nControlsAreaNewRightCoord = correctControlsAreaNewRightCoord( nControlsAreaNewRightCoord );

    long nControlsAreaOldRightCoord = m_rectMsg.right;
    //long cxControlsArea = nControlsAreaNewRightCoord - m_ptControlsAreaTopLeft.x;
    long dx;
    int  i;

    m_rectMsg.right =
    m_rectLongMsg.right = nControlsAreaNewRightCoord;

    for( i = 0 ; i < MaxCheckboxes ; ++i )
        m_arectChecks[i].right = nControlsAreaNewRightCoord;

    // Align buttons area. (left, center or right)
    dx = 0;
    RECT rectBtns = getAllButtonsBoundsRect();
    switch( m_dwMmbFlags & MMB_Mask_ButtonAlignX )
    {
        case MMB_AlignButtonsLeft     : dx = 0; break;
        case MMB_AlignButtonsRight    : dx = nControlsAreaNewRightCoord - rectBtns.right; break;
        case MMB_AlignButtonsCentered : dx = ((nControlsAreaNewRightCoord - rectBtns.right) - (rectBtns.left - m_ptControlsAreaTopLeft.x)) / 2; break;
    }
    for( i = 0 ; i < MaxButtons ; ++i )
        OffsetRect( &m_arectBtns[i], dx, 0 );
    OffsetRect( &m_rectHelpBtn, dx, 0 );
    OffsetRect( &m_rectMore, dx, 0 );

    m_rectHr.right += (nControlsAreaNewRightCoord - nControlsAreaOldRightCoord);
    m_rectExtraMsg.right =
    m_rectExtraLink.right = nControlsAreaNewRightCoord;

    // Wnd
    dx = nControlsAreaNewRightCoord - nControlsAreaOldRightCoord;
    m_rectWnd.right += dx;
}

void MoreMessageBox::Controls::adjustHeightFromMsgs( long cyNewHeaderMsg, long cyNewMsg, long cyNewExtraMsg, long cyNewExtraLink )
{
    long dy, dyfull;    // used just to offset each icon


    // cyNewHeaderMsg

    dy = cyNewHeaderMsg - (m_rectHeaderMsg.bottom - m_rectHeaderMsg.top);

    m_rectHeaderMsg.bottom += dy;
    offsetControls( 0, dy, true, false, true, true, true, true, true, true, true );


    // cyNewMsg

    dy = cyNewMsg - (m_rectMsg.bottom - m_rectMsg.top);

    // if we have an icon, the height must not become smaller than the icon
    if( dy > 0 || !m_hwndMsgIcon )
    {
        m_rectMsg.bottom += dy;
        offsetControls( 0, dy, true, false, false, false, true, true, true, true, true );
    }

    // if we have an icon and the new height wants to be smaller,
    // we align the text vertically to the icon as desired.
    if( dy < 0 && m_hwndMsgIcon )
    {
        m_rectMsg.bottom += dy;
        long dy2 = 0;
        switch( m_dwMmbFlags & MMB_Mask_MsgTextAlignY )
        {
            case MMB_AlignMsgTextTop       : dy2 = 0; break;
            case MMB_AlignMsgTextVCentered : dy2 = -dy / 2; break;
            case MMB_AlignMsgTextBottom    : dy2 = -dy; break;
        }
        OffsetRect( &m_rectMsg, 0, dy2 );
    }

    // Now we position the MsgIcon.
    if( dy > 0 )
    {
        dyfull = cyNewMsg - (m_rectMsgIcon.bottom - m_rectMsgIcon.top);
        long dy = 0;    // offset. this time measured from top of new Msg rect (which is same top as the old Msg rect).
        switch( m_dwMmbFlags & MMB_Mask_MsgIconAlignY )
        {
            case MMB_AlignMsgIconTop       : dy = 0; break;
            case MMB_AlignMsgIconVCentered : dy = dyfull / 2; break;
            case MMB_AlignMsgIconBottom    : dy = dyfull; break;
        }
        if( m_dwMmbFlags & MMB_MsgIcon_ExtraOffsetY )
            dy += (m_rectMsgIcon.bottom - m_rectMsgIcon.top) / 2;
        OffsetRect( &m_rectMsgIcon, 0, dy - (m_rectMsgIcon.top - m_rectMsg.top) );
        if( m_rectMsgIcon.bottom > m_rectMsg.bottom ) OffsetRect( &m_rectMsgIcon, 0, m_rectMsg.bottom - m_rectMsgIcon.bottom );
        if( m_rectMsgIcon.top    < m_rectMsg.top    ) OffsetRect( &m_rectMsgIcon, 0, m_rectMsg.top    - m_rectMsgIcon.top    );
    }


    // cyNewExtraMsg and cyNewExtraLink

    long cyNewExtra = 0;
    if( m_hwndExtraMsg  ) cyNewExtra += cyNewExtraMsg;
    if( m_hwndExtraLink ) cyNewExtra += cyNewExtraLink;
    if( m_hwndExtraMsg && m_hwndExtraLink ) cyNewExtra += m_dyExtraLinkDist;
    if( m_hwndExtraIcon )
        if( cyNewExtra < (m_rectExtraIcon.bottom - m_rectExtraIcon.top) )
            cyNewExtra = (m_rectExtraIcon.bottom - m_rectExtraIcon.top);

    RECT rectOldExtra = { m_rectWnd.right - m_rectWnd.left, m_rectWnd.bottom - m_rectWnd.top, 0, 0 };
    if( m_hwndExtraMsg  ) rectOldExtra = m_rectExtraMsg;
    if( m_hwndExtraLink ) { RECT rTmp = rectOldExtra; UnionRect( &rectOldExtra, &rTmp, &m_rectExtraLink ); }
    if( m_hwndExtraIcon ) { RECT rTmp = rectOldExtra; UnionRect( &rectOldExtra, &rTmp, &m_rectExtraIcon ); }
    long cyOldExtra = rectOldExtra.bottom - rectOldExtra.top;

    dy = cyNewExtra - cyOldExtra;

    if( cyOldExtra > 0 )
    {
        m_rectExtraMsg.bottom = m_rectExtraMsg.top + cyNewExtraMsg;
        m_rectExtraLink.top = m_rectExtraMsg.bottom + m_dyExtraLinkDist;
        m_rectExtraLink.bottom = m_rectExtraLink.top + cyNewExtraLink;

        m_rectWnd.bottom += dy;
    }

    // Finally we position the ExtraIcon.
    dyfull = cyNewExtra - (m_rectExtraIcon.bottom - m_rectExtraIcon.top);
    dy = 0;     // offset. this time measured from top of new Extra area (which is same top as the old Extra area).
    switch( m_dwMmbFlags & MMB_Mask_ExtraIconAlignY )
    {
        case MMB_AlignExtraIconTop       : dy = 0; break;
        case MMB_AlignExtraIconVCentered : dy = dyfull / 2; break;
        case MMB_AlignExtraIconBottom    : dy = dyfull; break;
    }
    OffsetRect( &m_rectExtraIcon, 0, dy - (m_rectExtraIcon.top - rectOldExtra.top) );
}

void MoreMessageBox::Controls::offsetControls( long dx, long dy,
                                               bool bDialogSize, bool bHeader,
                                               bool bMsgIcon, bool bMsg, bool bLongMsg,
                                               bool bButtons, bool bCheckboxes,
                                               bool bHorzRuler, bool bExtraArea )
{
    if( dx == 0 && dy == 0 ) return;

    if( bHeader )
        OffsetRect( &m_rectHeaderMsg, dx, dy );

    if( bMsgIcon )
        OffsetRect( &m_rectMsgIcon, dx, dy );

    if( bMsg )
        OffsetRect( &m_rectMsg, dx, dy );

    if( bLongMsg )
        OffsetRect( &m_rectLongMsg, dx, dy );

    if( bButtons )
    {
        for( int i = 0 ; i < MaxButtons ; ++i )
            OffsetRect( &m_arectBtns[i], dx, dy );
        OffsetRect( &m_rectHelpBtn, dx, dy );
        OffsetRect( &m_rectMore, dx, dy );
    }

    if( bCheckboxes )
    {
        for( int i = 0 ; i < MaxCheckboxes ; ++i )
            OffsetRect( &m_arectChecks[i], dx, dy );
    }

    if( bHorzRuler )
        OffsetRect( &m_rectHr, dx, dy );

    if( bExtraArea )
    {
        OffsetRect( &m_rectExtraIcon, dx, dy );
        OffsetRect( &m_rectExtraMsg, dx, dy );
        OffsetRect( &m_rectExtraLink, dx, dy );
    }

    if( bDialogSize )
    {
        m_rectWnd.right  += dx;
        m_rectWnd.bottom += dy;
    }
}

void MoreMessageBox::Controls::inflateBorders( RECT rectInflate )
{
    offsetControls( -rectInflate.left, -rectInflate.top,
                    false, true, true, true, true, true, true, true, true );

    m_rectWnd.right  -= rectInflate.left + rectInflate.right;
    m_rectWnd.bottom -= rectInflate.top  + rectInflate.bottom;
}

void MoreMessageBox::Controls::decreaseNumCheckboxesTo( int cCheckboxes )
{
    for( int i = MaxCheckboxes - 1 ; i >= 1 && cCheckboxes <= i ; --i )
        if( m_ahwndChecks[i] )
        {
            DestroyWindow( m_ahwndChecks[i] );
            m_ahwndChecks[i] = NULL;

            offsetControls( 0, -m_dyCheckboxes, true, false, false, false, false, false, false, true, true );
        }

    if( cCheckboxes <= 0 && m_ahwndChecks[0] )
    {
        DestroyWindow( m_ahwndChecks[0] );
        m_ahwndChecks[0] = NULL;

        offsetControls( 0, -m_dyCheck1, true, false, false, false, false, false, false, true, true );
    }
}

///<summary>
///  <b>setRects()</b> sets the positions and sizes of
///  all controls according to the current rects.
///</summary>
void MoreMessageBox::Controls::setRects( bool bKeepDialogVisible )
{
    int i;

    SetControlRect( m_hwndHeaderMsg, m_rectHeaderMsg );

    SetControlTopLeft( m_hwndMsgIcon, m_rectMsgIcon );

    SetControlRect( m_hwndMsg, m_rectMsg );
    SetControlRect( m_hwndLongMsg, m_rectLongMsg );

    for( i = 0 ; i < MaxButtons ; ++i )
        SetControlRect( m_ahwndBtns[i], m_arectBtns[i] );
    SetControlRect( m_hwndHelpBtn, m_rectHelpBtn );
    SetControlRect( m_hwndMore, m_rectMore );

    for( i = 0 ; i < MaxCheckboxes ; ++i )
        SetControlRect( m_ahwndChecks[i], m_arectChecks[i] );

    SetControlRect( m_hwndHr, m_rectHr );

    SetControlTopLeft( m_hwndExtraIcon, m_rectExtraIcon );
    SetControlRect( m_hwndExtraMsg, m_rectExtraMsg );
    SetControlRect( m_hwndExtraLink, m_rectExtraLink );

    if( m_hWnd )
        SetWindowPos( m_hWnd, HWND_TOP,
                      m_rectWnd.left, m_rectWnd.top,
                      m_rectWnd.right - m_rectWnd.left, m_rectWnd.bottom - m_rectWnd.top,
                      SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );

    if( m_hWnd && bKeepDialogVisible )
    {
        // Now we ensure that the entire dialog fits within the desktop area.
        // The dialog will be eventually repositioned if necessary.
        SendMessage( m_hWnd, DM_REPOSITION, 0, 0 );
    }
}

void MoreMessageBox::Controls::SetControlRect( HWND hwndCtrl, RECT rect, bool bDisableIfEmptyRect )
{
    if( hwndCtrl )
    {
        SetWindowPos( hwndCtrl, HWND_TOP,
                      rect.left, rect.top,
                      rect.right - rect.left, rect.bottom - rect.top,
                      SWP_NOZORDER | SWP_NOACTIVATE );

        if( bDisableIfEmptyRect && IsRectEmpty(&rect) )
            EnableWindow( hwndCtrl, FALSE );
    }
}

void MoreMessageBox::Controls::SetControlTopLeft( HWND hwndCtrl, POINT pt )
{
    if( hwndCtrl )
    {
        SetWindowPos( hwndCtrl, HWND_TOP,
                      pt.x, pt.y,
                      0, 0,
                      SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
    }
}

///<remarks> Do not do calculations of the other methods in here
///  when the dialog is collapsed. This could cause undesired behaviour.
///  Better to expand it first. </remarks>
void MoreMessageBox::Controls::Collapse()
{
    if( !m_bIsExpanded ) return;
    if( !m_hwndLongMsg ) return;

    long dy = -m_dyExpand;
    offsetControls( 0, dy, true, false, false, false, false, true, true, true, true );

    if( m_hwndLongMsg )
    {
        EnableWindow( m_hwndLongMsg, FALSE );
        ShowWindow( m_hwndLongMsg, SW_HIDE );
    }

    if( m_hwndMore )
        EnableWindow( m_hwndMore, TRUE );

    m_bIsExpanded = false;
}

///<remarks> Do not do calculations of the other methods in here
///  when the dialog is collapsed. This could cause undesired behaviour.
///  Better to expand it first. </remarks>
void MoreMessageBox::Controls::Expand()
{
    if( m_bIsExpanded ) return;
    if( !m_hwndLongMsg ) return;

    long dy = m_dyExpand;
    offsetControls( 0, dy, true, false, false, false, false, true, true, true, true );

    if( m_hwndLongMsg )
    {
        EnableWindow( m_hwndLongMsg, TRUE );
        ShowWindow( m_hwndLongMsg, SW_SHOWNA );
    }

    if( m_hwndMore && (m_dwMmbFlags & MMB_ExpandableOnly) )
        EnableWindow( m_hwndMore, FALSE );

    m_bIsExpanded = true;
}

#endif




}
