/*
    File:               argonRandom.h
    Description:        provides Random number generators
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         05.11.2006
    last updated:       05.11.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonRandom_h
#define __argonRandom_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonArray.h"
#endif

namespace ArgonLibrary
{




///<summary> This enum describes how the init params are set. </summary>
enum eMersenneTwisterKind
{
    mersenneKind_default        = 0,    // this is my adopted kind to have best handling I could find out so far
    mersenneKind_common                 // this is the original mersenne twister handling
};

///<summary> This is a pseudo random number generator. </summary>
class MersenneTwister
{
    protected :
        int   m_n, m_m;
        DWORD m_matrix_a, m_upper_mask, m_lower_mask, m_tempering_mask_b, m_tempering_mask_c;
        bool  m_swap_mag01;
        DWORD m_seed_1, m_seed_2, m_seed_3, m_seed_4;

        DWORD m_good_seed;

        Array<DWORD> m_mt;
        int          m_mti;     // if m_mti == m_n+1 means 'm_mt' is not initialized
        DWORD        m_mag01[2];

    public :
        MersenneTwister() { m_good_seed = 4357; setSeed( m_good_seed ); }
        MersenneTwister( DWORD seed ) { m_good_seed = 4357; setSeed( seed ); }

        void  setGoodSeed( DWORD good_seed ) { m_good_seed = good_seed; }
        DWORD getGoodSeed() const { return m_good_seed; }

        void setSeed() { setSeed( m_good_seed ); }
        void setSeed( DWORD seed ) { setSeed( seed, mersenneKind_default ); }
        bool setSeed( DWORD seed, eMersenneTwisterKind kind );
        void setSeed( DWORD seed,
                      int n, int m,
                      DWORD matrix_a,
                      DWORD upper_mask, DWORD lower_mask,
                      DWORD tempering_mask_b, DWORD tempering_mask_c,
                      bool swap_mag01,
                      DWORD seed_1, DWORD seed_2, DWORD seed_3, DWORD seed_4 );
        DWORD  nextDWord();
        WORD   nextWord()  { return (WORD)( nextDWord() >> 16 ); }
        BYTE   nextByte()  { return (BYTE)( nextDWord() >> 24 ); }
        int    nextInt()   { return (int)   nextDWord();         }
        short  nextShort() { return (short) nextWord();          }
        char   nextChar()  { return (char)( nextDWord() >> 16 ); }
        bool   nextBool()  { return (bool)( nextInt() < 0 );     }
        float  nextFloat();
        double nextDouble();
};




}

#endif
